import numpy as np
import torch


class DummyEncoder(torch.nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dummy = torch.nn.Linear(1, 1)
        self.cat_names = []
        self.num_values = []
        self.cat_max_size = {}
        self.embed_size = {}

    def forward(self, x):
        return x

    @property
    def category_names(self):
        return set(self.cat_names + self.num_values)

    @property
    def category_max_size(self):
        return self.cat_max_size



import torch
import torch.nn as nn
import torch.nn.functional as F
from numpy.testing import assert_almost_equal

from .metric import outer_cosine_similarity


class ContrastiveLoss(nn.Module):
    """
    Contrastive loss
    
    "Signature verification using a siamese time delay neural network", NIPS 1993
    https://papers.nips.cc/paper/769-signature-verification-using-a-siamese-time-delay-neural-network.pdf
    """

    def __init__(self, margin, pair_selector):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin
        self.pair_selector = pair_selector

    def forward(self, embeddings, target):
        
        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)
        positive_loss = F.pairwise_distance(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]]).pow(2)
        
        negative_loss = F.relu(
            self.margin - F.pairwise_distance(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]])
        ).pow(2)
        loss = torch.cat([positive_loss, negative_loss], dim=0)
        
        return loss.sum(), len(positive_pairs) + len(negative_pairs)


class BinomialDevianceLoss(nn.Module):
    """
    Binomial Deviance loss

    "Deep Metric Learning for Person Re-Identification", ICPR2014
    https://arxiv.org/abs/1407.4979
    """

    def __init__(self, pair_selector, alpha=1, beta=1, C=1):
        super(BinomialDevianceLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.C = C
        self.pair_selector = pair_selector

    def forward(self, embeddings, target):
        
        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)
        
        pos_pair_similarity = F.cosine_similarity(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]], dim=1)
        neg_pair_similarity = F.cosine_similarity(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]], dim=1)

        pos_loss = torch.mean(torch.log(1 + torch.exp(-self.alpha*(pos_pair_similarity - self.beta))))
        neg_loss = torch.mean(torch.log(1 + torch.exp(self.alpha*self.C*(neg_pair_similarity - self.beta))))

        res_loss = (pos_loss + neg_loss) * (len(target))
        
        return res_loss, len(positive_pairs) + len(negative_pairs)


class TripletLoss(nn.Module):
    """
    Triplets loss

    "Deep metric learning using triplet network", SIMBAD 2015
    https://arxiv.org/abs/1412.6622
    """

    def __init__(self, margin, triplet_selector):
        super(TripletLoss, self).__init__()
        self.margin = margin
        self.triplet_selector = triplet_selector

    def forward(self, embeddings, target):

        triplets = self.triplet_selector.get_triplets(embeddings, target)

        if embeddings.is_cuda:
            triplets = triplets.cuda()

        ap_distances = F.pairwise_distance(embeddings[triplets[:, 0]], embeddings[triplets[:, 1]])
        an_distances = F.pairwise_distance(embeddings[triplets[:, 0]], embeddings[triplets[:, 2]])
        losses = F.relu(ap_distances - an_distances + self.margin)

        return losses.sum(), len(triplets)


class HistogramLoss(torch.nn.Module):
    """
    HistogramLoss
    
    "Learning deep embeddings with histogram loss", NIPS 2016
    https://arxiv.org/abs/1611.00822
    code based on https://github.com/valerystrizh/pytorch-histogram-loss
    """
    
    def __init__(self, num_steps=100):
        super(HistogramLoss, self).__init__()
        self.step = 2 / (num_steps - 1)
        self.eps = 1 / num_steps
        self.t = torch.arange(-1, 1+self.step, self.step).view(-1, 1)
        self.tsize = self.t.size()[0]
        self.device = None
        
    def forward(self, embeddings, classes):
        def histogram(inds, size):
            
            s_repeat_ = s_repeat.clone()
            indsa = (s_repeat_floor - (self.t - self.step) > -self.eps) &  \
                    (s_repeat_floor - (self.t - self.step) < self.eps) & inds

            assert indsa.nonzero(as_tuple=False).size()[0] == size, 'Another number of bins should be used'
            zeros = torch.zeros((1, indsa.size()[1])).bool()
            zeros = zeros.to(self.device)
            indsb = torch.cat((indsa, zeros))[1:, :]
            s_repeat_[~(indsb | indsa)] = 0
            # indsa corresponds to the first condition of the second equation of the paper
            s_repeat_[indsa] = (s_repeat_ - self.t + self.step)[indsa] / self.step
            # indsb corresponds to the second condition of the second equation of the paper
            s_repeat_[indsb] = (-s_repeat_ + self.t + self.step)[indsb] / self.step

            return s_repeat_.sum(1) / size

        self.device = embeddings.device
        self.t = self.t.to(self.device)

        # L2 normalization
        classes_size = classes.size()[0]
        classes_eq = (classes.repeat(classes_size, 1)  == classes.view(-1, 1).repeat(1, classes_size)).data
        dists = outer_cosine_similarity(embeddings)
        
        assert ((dists > 1 + self.eps).sum().item() + (dists < -1 - self.eps).sum().item()) == 0, 'L2 normalization ' \
                                                                                                  'should be used '
        s_inds = torch.triu(torch.ones(classes_eq.size()), 1).bool()
        s_inds = s_inds.to(self.device)
        pos_inds = classes_eq[s_inds].repeat(self.tsize, 1)
        neg_inds = ~classes_eq[s_inds].repeat(self.tsize, 1)
        pos_size = classes_eq[s_inds].sum().item()
        neg_size = (~classes_eq[s_inds]).sum().item()
        s = dists[s_inds].view(1, -1)
        s = s.clamp(-1 + 1e-6, 1 - 1e-6)
        s_repeat = s.repeat(self.tsize, 1)
        s_repeat_floor = (torch.floor((s_repeat.data + 1.0 - 1e-6) / self.step) * self.step - 1.0).float()
        
        histogram_pos = histogram(pos_inds, pos_size)
        assert_almost_equal(histogram_pos.sum().item(), 1, decimal=1, 
                            err_msg='Not good positive histogram', verbose=True)
        histogram_neg = histogram(neg_inds, neg_size)
        assert_almost_equal(histogram_neg.sum().item(), 1, decimal=1, 
                            err_msg='Not good negative histogram', verbose=True)
        histogram_pos_repeat = histogram_pos.view(-1, 1).repeat(1, histogram_pos.size()[0])
        histogram_pos_inds = torch.tril(torch.ones(histogram_pos_repeat.size()), -1).bool()
        histogram_pos_inds = histogram_pos_inds.to(self.device)
        histogram_pos_repeat[histogram_pos_inds] = 0
        histogram_pos_cdf = histogram_pos_repeat.sum(0)
        loss = torch.sum(histogram_neg * histogram_pos_cdf)
        
        return loss, pos_size + neg_size


class MarginLoss(torch.nn.Module):
    
    """
    Margin loss
    
    "Sampling Matters in Deep Embedding Learning", ICCV 2017
    https://arxiv.org/abs/1706.07567

    """

    def __init__(self, pair_selector, margin=1, beta=1.2):
        super(MarginLoss, self).__init__()
        self.margin = margin
        self.beta = beta
        self.pair_selector = pair_selector
        
    def forward(self, embeddings, target):

        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)
        
        d_ap = F.pairwise_distance(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]])
        d_an = F.pairwise_distance(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]])

        pos_loss = torch.clamp(d_ap - self.beta + self.margin, min=0.0)
        neg_loss = torch.clamp(self.beta - d_an + self.margin, min=0.0)

        loss = torch.cat([pos_loss, neg_loss], dim=0)
        
        return loss.sum(), len(positive_pairs) + len(negative_pairs)


class ComplexLoss(torch.nn.Module):
    def __init__(self, ml_loss, aug_loss, ml_loss_weight=1.):
        super(ComplexLoss, self).__init__()
        self.aug_loss = aug_loss
        self.ml_loss = ml_loss
        self.ml_loss_weight = ml_loss_weight

    def forward(self, model_outputs, target):
        aug_output, ml_output = model_outputs
        aug_target = target[:, 0]
        ml_target = target[:, 1]
        aug = self.aug_loss(aug_output, aug_target) * (1 - self.ml_loss_weight)
        ml = self.ml_loss(ml_output, ml_target) * self.ml_loss_weight
        return aug + ml


class BarlowTwinsLoss(torch.nn.Module):
    """
    From https://github.com/facebookresearch/barlowtwins

    """
    def __init__(self, lambd):
        super().__init__()

        self.lambd = lambd

    def forward(self, model_outputs, target):
        n = len(model_outputs)
        ix1 = torch.arange(0, n, 2, device=model_outputs.device)
        ix2 = torch.arange(1, n, 2, device=model_outputs.device)

        assert (target[ix1] == target[ix2]).all(), "Wrong embedding positions"

        z1 = model_outputs[ix1]
        z2 = model_outputs[ix2]

        # empirical cross-correlation matrix
        c = torch.mm(z1.T, z2)
        c.div_(n // 2)

        on_diag = torch.diagonal(c).add_(-1).pow_(2).sum()
        off_diag = self.off_diagonal(c).pow_(2).sum()
        loss = on_diag + self.lambd * off_diag
        return loss, None

    @staticmethod
    def off_diagonal(x):
        # return a flattened view of the off-diagonal elements of a square matrix
        n, m = x.shape
        assert n == m
        return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


def get_loss(params, sampling_strategy, kw_params=None):

    if params['train.loss'] == 'ContrastiveLoss':
        kwargs = {
            'margin': params.get('train.margin', None),
            'pair_selector': sampling_strategy
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = ContrastiveLoss(**kwargs)

    elif params['train.loss'] == 'BinomialDevianceLoss':
        kwargs = {
            'C': params.get('train.C', None),
            'alpha': params.get('train.alpha', None),
            'beta': params.get('train.beta', None),
            'pair_selector': sampling_strategy
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = BinomialDevianceLoss(**kwargs)       

    elif params['train.loss'] == 'TripletLoss':
        kwargs = {
            'margin': params.get('train.margin', None),
            'triplet_selector': sampling_strategy
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = TripletLoss(**kwargs)

    elif params['train.loss'] == 'HistogramLoss':
        kwargs = {
            'num_steps': params.get('train.num_steps', None),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = HistogramLoss(**kwargs)

    elif params['train.loss'] == 'MarginLoss':
        kwargs = {
            'margin': params.get('train.margin', None),
            'beta': params.get('train.beta', None),
            'pair_selector': sampling_strategy
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = MarginLoss(**kwargs)
    elif params['train.loss'] == 'CPCLoss':
        kwargs = {
            'k_pos_samples': params.get('cpc.k_pos_samples', None),
            'm_neg_samples': params.get('cpc.m_neg_samples', None),
            'linear_predictor': kw_params['linear_predictor'],
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
    elif params['train.loss'] == 'BarlowTwinsLoss':
        kwargs = {
            'lambd': params.get('train.lambd', None),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        loss_fn = BarlowTwinsLoss(**kwargs)
    else:
        raise AttributeError(f'wrong loss "{params["train.loss"]}"')

    def loss(*args, **kwargs):
        return loss_fn(*args, **kwargs)[0]

    return loss



import torch
from pyhocon import ConfigFactory

from dltranz.seq_encoder.agg_feature_model import AggFeatureModel
from dltranz.trx_encoder import PaddedBatch


def get_data():
    return PaddedBatch(
        payload={
            'cat1': torch.tensor([
                [1, 3, 5, 0, 0, 0],
                [2, 5, 3, 1, 1, 0],
                [4, 0, 0, 0, 0, 0],
            ]),
            'num1': torch.tensor([
                [5, 6, 3, 0, 0, 0],
                [4, 0, 1, 3, 2, 0],
                [7, 0, 0, 0, 0, 0],
            ]).float(),
        },
        length=torch.tensor([3, 5, 1]),
    )


def get_conf():
    params = {
        'params': {
            'trx_encoder': {
                'numeric_values': {
                    'num1': {},
                },
                'embeddings': {
                    'cat1': {'in': 8}
                },
                'was_logified': False,
                'log_scale_factor': 1.0,
                "num_aggregators": {
                    "count": False,
                    "sum": True,
                    "std": True
                },
                "cat_aggregators": {
                    "top_k": 3
                }
            }
        }
    }
    return ConfigFactory.from_dict(params)


def test_output_size():
    out_size = AggFeatureModel(get_conf()['params.trx_encoder']).output_size
    assert out_size == 24


def test_model():
    model = AggFeatureModel(get_conf()['params.trx_encoder'])
    out = model(get_data())
    assert out.size() == (3, 24)



import torch

from dltranz.trx_encoder import NoisyEmbedding


def test_no_noise():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=0.0)
    x = torch.zeros(4, 8, dtype=torch.long)
    out = embedding(x)
    assert out.size() == (4, 8, 4)


def test_noisy_embedding():
    embedding = NoisyEmbedding(16, 4, 0, noise_scale=1.0)
    x = torch.zeros(4, 8, dtype=torch.long)
    out = embedding(x)
    assert out.size() == (4, 8, 4)



import numpy as np
import torch
from torch import nn as nn

from dltranz.seq_encoder.abs_seq_encoder import AbsSeqEncoder
from dltranz.seq_encoder.utils import LastStepEncoder
from dltranz.trx_encoder import PaddedBatch, TrxEncoder


class RnnEncoder(nn.Module):
    def __init__(self, input_size, config):
        super().__init__()

        self.hidden_size = config['hidden_size']
        self.rnn_type = config['type']
        self.bidirectional = config['bidir']
        if self.bidirectional:
            raise AttributeError('bidirectional RNN is not supported yet')
        self.trainable_starter = config['trainable_starter']

        # initialize RNN
        if self.rnn_type == 'lstm':
            self.rnn = nn.LSTM(
                input_size,
                self.hidden_size,
                num_layers=1,
                batch_first=True,
                bidirectional=self.bidirectional)
        elif self.rnn_type == 'gru':
            self.rnn = nn.GRU(
                input_size,
                self.hidden_size,
                num_layers=1,
                batch_first=True,
                bidirectional=self.bidirectional)
        else:
            raise Exception(f'wrong rnn type "{self.rnn_type}"')

        self.full_hidden_size = self.hidden_size if not self.bidirectional else self.hidden_size * 2

        # initialize starter position if needed
        if self.trainable_starter == 'static':
            num_dir = 2 if self.bidirectional else 1
            self.starter_h = nn.Parameter(torch.randn(num_dir, 1, self.hidden_size))

    def forward(self, x: PaddedBatch, h_0: torch.Tensor = None):
        """

        :param x:
        :param h_0: None or [1, B, H] float tensor
                    0.0 values in all components of hidden state of specific client means no-previous state and
                    use starter for this client
                    h_0 = None means no-previous state for all clients in batch
        :return:
        """
        shape = x.payload.size()
        assert shape[1] > 0, "Batch can'not have 0 transactions"

        # prepare initial state
        if self.trainable_starter == 'static':
            starter_h = self.starter_h.expand(-1, shape[0], -1).contiguous()
            if h_0 is None:
                h_0 = starter_h
            elif h_0 is not None and not self.training:
                h_0 = torch.where(
                    (h_0.squeeze(0).abs().sum(dim=1) == 0.0).unsqueeze(0).unsqueeze(2).expand(*starter_h.size()),
                    starter_h,
                    h_0,
                )
            else:
                raise NotImplementedError('Unsupported mode: cannot mix fixed X and learning Starter')

        # pass-through rnn
        if self.rnn_type == 'lstm':
            out, _ = self.rnn(x.payload)
        elif self.rnn_type == 'gru':
            out, _ = self.rnn(x.payload, h_0)
        else:
            raise Exception(f'wrong rnn type "{self.rnn_type}"')

        return PaddedBatch(out, x.seq_lens)


class RnnSeqEncoder(AbsSeqEncoder):
    def __init__(self, params, is_reduce_sequence):
        super().__init__(params, is_reduce_sequence)

        p = TrxEncoder(params['trx_encoder'])
        e = RnnEncoder(p.output_size, params['rnn'])
        layers = [p, e]
        self.reducer = LastStepEncoder()
        self.model = torch.nn.Sequential(*layers)

    @property
    def category_max_size(self):
        return self.model[0].category_max_size

    @property
    def category_names(self):
        return self.model[0].category_names

    @property
    def embedding_size(self):
        return self.params['rnn']['hidden_size']

    def forward(self, x):
        x = self.model(x)
        if self.is_reduce_sequence:
            x = self.reducer(x)
        return x


class RnnInference(torch.nn.Module):
    def __init__(self, model: RnnSeqEncoder):
        super().__init__()

        self.hidden_size = hidden_size
        self.is_reduce_sequence = model.seq_encoder.is_reduce_sequence
        self.p = model.seq_encoder.model[0]
        self.e = model.seq_encoder.model[1]
        self.reducer = model.seq_encoder.reducer

    def forward(self, x, h_0=None):
        x_ = self.p(x)
        x = self.e(x_, h_0)
        if self.is_reduce_sequence:
            x = self.reducer(x)
        return x


class RnnSeqEncoderDistributionTarget(RnnSeqEncoder):
    def transform(self, x):
        return np.sign(x) * np.log(np.abs(x) + 1)

    def transform_inv(self, x):
        return np.sign(x) * (np.exp(np.abs(x)) - 1)

    def __init__(self, params, is_reduce_sequence):
        super().__init__(params, is_reduce_sequence)
        head_params = dict(params['head_layers']).get('CombinedTargetHeadFromRnn', None)
        self.pass_samples = head_params.get('pass_samples', True)
        self.numeric_name = list(params['trx_encoder']['numeric_values'].keys())[0]
        self.collect_pos, self.collect_neg = (head_params.get('pos', True), head_params.get('neg', True)) if head_params else (0, 0)
        self.eps = 1e-7

    def forward(self, x):
        amount_col = []
        for i, row in enumerate(x.payload[self.numeric_name]):
            amount_col += [list(row[:x.seq_lens[i].item()].cpu().numpy())]
        amount_col = np.array(amount_col, dtype=object)
        neg_sums = []
        pos_sums = []
        for list_row in amount_col:
            np_row = np.array(list_row)
            if self.collect_neg:
                neg_sums += [np.sum(self.transform_inv(np_row[np.where(np_row < 0)]))]
            if self.collect_pos:
                pos_sums += [np.sum(self.transform_inv(np_row[np.where(np_row >= 0)]))]
        neg_sum_logs = np.log(np.abs(np.array(neg_sums)) + self.eps)
        pos_sum_logs = np.log(np.array(pos_sums) + self.eps)

        x = super().forward(x)
        if (not self.pass_samples):
            return x
        if self.collect_neg and self.collect_pos:
            return x, neg_sum_logs, pos_sum_logs
        elif self.collect_neg:
            return x, neg_sum_logs
        elif self.collect_pos:
            return x, pos_sum_logs


class SkipStepEncoder(nn.Module):
    def __init__(self, step_size):
        super().__init__()
        self.step_size = step_size

    def forward(self, x: PaddedBatch):
        max_len = x.payload.shape[1] - 1
        s = self.step_size

        first_dim_idx = []
        second_dim_idx = []
        for i, l in enumerate(x.seq_lens):
            idx_to_take = np.arange(min(l - 1, s - 1 + l % s), l, s)
            pad_idx = np.array([max_len - 1] * (max_len // s - len(idx_to_take)), dtype=np.int32)
            idx_to_take = np.concatenate([[-1], idx_to_take, pad_idx]) + 1
            first_dim_idx.append(np.ones(len(idx_to_take)) * i)
            second_dim_idx.append(idx_to_take)

        out = x.payload[first_dim_idx, second_dim_idx]
        out_lens = torch.tensor([min(1, l // self.step_size) for l in x.seq_lens])

        return PaddedBatch(out, out_lens)


def skip_rnn_encoder(input_size, params):
    rnn0 = RnnEncoder(input_size, params['rnn0'])
    rnn1 = RnnEncoder(params['rnn0']['hidden_size'], params['rnn1'])
    sse = SkipStepEncoder(params['skip_step_size'])
    return nn.Sequential(rnn0, sse, rnn1)



import torch

from dltranz.train import get_optimizer


def test_get_optimizer():
    model = torch.nn.Linear(10, 1)
    params = {
        'train.lr': 0.01,
        'train.weight_decay': 0.01,
    }
    optim = get_optimizer(model, params)


def test_get_optimizer_params():
    model = torch.nn.Sequential(
        torch.nn.Linear(10, 5),
        torch.nn.Sigmoid(),
        torch.nn.Linear(5, 1),
        torch.nn.Sigmoid(),
    )
    print([k for k, v in model.named_parameters()])
    params = {
        'train.lr': 0.01,
        'train.weight_decay': 0.01,
        'train.optimiser_params': {
            '0.weight': {'lr': 0.001},
        },
    }

    optim = get_optimizer(model, params)
    for grp in optim.param_groups:
        if len(grp['params']) == 1:
            assert grp['lr'] == 0.001
        elif len(grp['params']) == 3:
            assert grp['lr'] == 0.01
        else:
            raise AssertionError('Error in test. Only 4 parameters in total')


def test_get_optimizer_2params():
    model = torch.nn.Sequential(
        torch.nn.Linear(10, 10),
        torch.nn.Sigmoid(),
        torch.nn.Linear(10, 5),
        torch.nn.Sigmoid(),
        torch.nn.Linear(5, 1),
        torch.nn.Sigmoid(),
    )
    print([k for k, v in model.named_parameters()])
    params = {
        'train.lr': 0.01,
        'train.weight_decay': 0.01,
        'train.optimiser_params': {
            '0.weight': {'lr': 0.001},
            '2.weight': {'lr': 0.001},
        },
    }

    optim = get_optimizer(model, params)
    for grp in optim.param_groups:
        if len(grp['params']) == 1:
            assert grp['lr'] == 0.001
        elif len(grp['params']) == 4:
            assert grp['lr'] == 0.01
        else:
            raise AssertionError('Error in test. Only 6 parameters in total')



import torch

from dltranz.lightning_modules.AbsModule import ABSModule
from dltranz.metric_learn.losses import get_loss
from dltranz.metric_learn.metric import BatchRecallTopPL
from dltranz.metric_learn.sampling_strategies import get_sampling_strategy


class EmbModule(ABSModule):
    """pl.LightningModule for training CoLES embeddings

    Parameters
    ----------
    seq_encoder : torch.nn.Module
        sequence encoder
    head : torch.nn.Module
        head of th model
    margin : float. Default: 0.5
        The margin for Contrastive Loss
    neg_count : int. Default: 5
        The number of negative samples for hard negative sampler
    lr : float. Default: 1e-3
        Learning rate
    weight_decay: float. Default: 0.0
        weight decay for optimizer
    lr_scheduler_step_size : int. Default: 100
        Period of learning rate decay.
    lr_scheduler_step_gamma: float. Default: 0.1
        Multiplicative factor of learning rate decay.
    """

    def __init__(self,
                 seq_encoder: torch.nn.Module,
                 head: torch.nn.Module,
                 margin: float = 0.5,
                 neg_count: int = 5,
                 lr: float = 1e-3,
                 weight_decay: float = 0.0,
                 lr_scheduler_step_size: int = 100,
                 lr_scheduler_step_gamma: float = 0.1):

        train_params = {
            'train.sampling_strategy': 'HardNegativePair',
            'train.loss': 'ContrastiveLoss',
            'train.margin': margin,
            'train.neg_count': neg_count,
            'train.lr': lr,
            'train.weight_decay': weight_decay,
            'lr_scheduler': {
                'step_size': lr_scheduler_step_size,
                'step_gamma': lr_scheduler_step_gamma
            }
        }
        super().__init__(train_params, seq_encoder)

        self._head = head

    @property
    def metric_name(self):
        return 'recall_top_k'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def get_loss(self):
        sampling_strategy = get_sampling_strategy(self.hparams.params)
        loss = get_loss(self.hparams.params, sampling_strategy)
        return loss

    def get_validation_metric(self):
        default_kwargs = {
            'K': 4,
            'metric': 'cosine'
        }
        if 'validation_metric_params' in self.hparams.params:
            kwargs = {**default_kwargs, **self.hparams.params['validation_metric_params']}
        else:
            kwargs = default_kwargs

        return BatchRecallTopPL(**kwargs)

    def shared_step(self, x, y):
        y_h = self(x)
        if self._head is not None:
            y_h = self._head(y_h)
        return y_h, y



from typing import List


class DataPreprocessor:
    def __init__(self,
                 col_id: str,
                 cols_event_time: str,
                 cols_category: List[str],
                 cols_log_norm: List[str],
                 cols_identity: List[str],
                 cols_target: List[str],
                 ):
        self.col_id = col_id
        self.cols_event_time = cols_event_time
        self.cols_category = cols_category
        self.cols_log_norm = cols_log_norm
        self.cols_identity = cols_identity
        self.cols_target = cols_target
        self._reset()

    def _reset(self):
        """Reset internal data-dependent state of the preprocessor, if necessary.
        __init__ parameters are not touched.
        """
        self.cols_category_mapping = {}
        self.cols_log_norm_maxes = {}

    def check_is_fitted(self):
        """Check is preprocessor is fitted, before applying transformation
        """
        assert all(k in self.cols_category_mapping for k in self.cols_category), "Preprocessor is not fitted"

    def fit(self, dt, **params):
        raise NotImplementedError()

    def transform(self, dt, **params):
        raise NotImplementedError()

    def fit_transform(self, dt, **fit_params):
        """
        Fit to data, then transform it.
        Fits transformer to `dt` with optional parameters `fit_params`
        and returns a transformed version of `dt`.
        Parameters
        ----------
        dt : Input samples.
        **fit_params : dict
            Additional fit parameters.
        Returns
        -------
        dt_transformed : Transformed data.
        """
        return self.fit(dt, **fit_params).transform(dt)

    def get_category_sizes(self):
        """Gets a dict of mapping to integers lengths for categories
        """
        return {k: len(v) for k, v in self.cols_category_mapping.items()}



import numpy as np
import torch
from pyhocon import ConfigFactory
from sklearn.metrics import roc_auc_score

from dltranz.data_load import create_validation_loader, TrxDataset
from dltranz.models import rnn_model
from .test_data_load import gen_trx_data
from dltranz.train import score_model, score_model


def tst_params():
    params = {
        'ensemble_size': 1,
        'score_metric': 'auroc',
        'norm_scores': True,
        'head': {
            "pred_all_states": False,
            "pred_all_states_mean": False,
            "explicit_lengths": True,
            "norm_input": False,
            'use_batch_norm': False,
        },
        "train": {
            "weight_decay": 0,
            "lr": 0.004,
            "loss": "bce",
            "random_neg": True,
            "batch_size": 32,
            "num_workers": 1,
            "trx_dropout": .1,
            "n_epoch": 1,
            "max_seq_len": 30
        },
        "valid": {
            "batch_size": 32,
            "num_workers": 1,
            "max_seq_len": 30
        },
        'rnn': {
            'trainable_starter': 'static',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        "trx_encoder": {
            "embeddings_noise": .1,
            "norm_embeddings": False,
            'embeddings': {
                'mcc_code': {'in': 21, 'out': 3},
                'trans_type': {'in': 11, 'out': 2},
            },
            'numeric_values': {'amount': 'log'}
        },
        "lr_scheduler": {
            "step_size": 10,
            "step_gamma": 0.8
        }
    }
    return ConfigFactory.from_dict(params)


def tst_params_skip_rnn():
    params = {
        'skip_rnn': {
            'rnn0': {
                'trainable_starter': 'static',
                'hidden_size': 16,
                'type': 'gru',
                'bidir': False,
            },
            'rnn1': {
                'trainable_starter': 'static',
                'hidden_size': 8,
                'type': 'gru',
                'bidir': False,
            },
            "skip_step_size": 3
        },
        "head": {
            "explicit_lengths": False,
        }
    }

    return ConfigFactory.from_dict(params).with_fallback(tst_params())


def tst_params_transf():
    params = {
        'transf': {
            'n_heads': 2,
            'dim_hidden': 16,
            'dropout': .1,
            'n_layers': 2,
            'max_seq_len': 200,
            'use_after_mask': True,
            'use_positional_encoding': True,
            'sum_output': True,
            'input_size': 16
        },
        "head": {
            "explicit_lengths": False,
        }
    }

    return ConfigFactory.from_dict(params).with_fallback(tst_params())


def tst_params_swa():
    params = {
        'train': {
            'swa': {
                'swa_start': 0
            }
        }
    }

    return ConfigFactory.from_dict(params).with_fallback(tst_params_skip_rnn())


def tst_loaders():
    train_data = gen_trx_data((torch.rand(1000) * 60 + 1).long())
    train_ds = TrxDataset(train_data)
    test_data = gen_trx_data((torch.rand(100) * 60 + 1).long())
    valid_ds = TrxDataset(test_data)
    return train_ds, valid_ds


def test_score_model():
    params = tst_params()

    test_data = gen_trx_data((torch.rand(1000)*60+1).long())
    valid_loader = create_validation_loader(TrxDataset(test_data), params['valid'])

    pred, true = score_model(rnn_model(params), valid_loader, params)
    print(roc_auc_score(true, pred))


def test_score_model_mult1():
    params = tst_params()

    test_data = gen_trx_data((torch.rand(1000)*60+1).long())
    valid_loader = create_validation_loader(TrxDataset(test_data), params['valid'])

    pred, true = score_model(rnn_model(params), valid_loader, params)
    print(roc_auc_score(true, pred))


def test_score_model_mult2():
    model = torch.nn.Sequential(torch.nn.Linear(16, 2), torch.nn.Sigmoid())
    valid_loader = [
        (torch.rand(4, 16), np.arange(4), np.arange(4).astype(str)),
        (torch.rand(2, 16), np.arange(2), np.arange(2).astype(str)),
        (torch.rand(1, 16), np.arange(1), np.arange(1).astype(str)),
    ]

    pred, id1, id2 = score_model(model, valid_loader, {'device': 'cpu'})
    assert pred.shape == (7, 2)
    assert id1.shape == (7,)
    assert id2.shape == (7,)

    np.testing.assert_array_almost_equal(id1, np.array([0, 1, 2, 3, 0, 1, 0]))
    assert id2.tolist() == ['0', '1', '2', '3', '0', '1', '0']



import torch

from torch.nn import Linear, BatchNorm1d, Sigmoid, Sequential, ReLU, Dropout, LogSoftmax
from typing import List

from dltranz.seq_encoder.rnn_encoder import RnnEncoder
from dltranz.seq_encoder.utils import PerTransHead, PerTransTransf, TimeStepShuffle, scoring_head
from dltranz.seq_encoder.rnn_encoder import skip_rnn_encoder
from dltranz.seq_encoder.transf_seq_encoder import TransformerSeqEncoder
from dltranz.trx_encoder import TrxEncoder, TrxMeanEncoder
from dltranz.custom_layers import Squeeze
from dltranz.seq_encoder.utils import NormEncoder


def trx_avg_model(params):
    p = TrxEncoder(params['trx_encoder'])
    h = PerTransHead(p.output_size)
    m = torch.nn.Sequential(p, h, torch.nn.Sigmoid())
    return m


def trx_avg2_model(params):
    p = TrxMeanEncoder(params['trx_encoder'])
    h = scoring_head(TrxMeanEncoder.output_size(params['trx_encoder']), params['head'])
    m = torch.nn.Sequential(p, h)
    return m


def rnn_model(params):
    p = TrxEncoder(params['trx_encoder'])
    e = RnnEncoder(p.output_size, params['rnn'])
    h = scoring_head(
        input_size=params['rnn.hidden_size'] * (2 if params['rnn.bidir'] else 1),
        params=params['head']
    )

    m = torch.nn.Sequential(p, e, h)
    return m


def rnn_shuffle_model(params):
    p = TrxEncoder(params['trx_encoder'])
    p_size = p.output_size
    p = torch.nn.Sequential(p, TimeStepShuffle())
    e = RnnEncoder(p_size, params['rnn'])
    h = scoring_head(
        input_size=params['rnn.hidden_size'] * (2 if params['rnn.bidir'] else 1),
        params=params['head']
    )

    m = torch.nn.Sequential(p, e, h)
    return m


def skip_rnn2_model(params):
    p = TrxEncoder(params['trx_encoder'])
    e = skip_rnn_encoder(p.output_size, params['skip_rnn'])
    h = scoring_head(
        input_size=params['skip_rnn.rnn1.hidden_size'] * (2 if params['skip_rnn.rnn1.bidir'] else 1),
        params=params['head']
    )

    m = torch.nn.Sequential(p, e, h)
    return m


def transformer_model(params):
    p = TrxEncoder(params['trx_encoder'])
    trx_size = p.output_size
    enc_input_size = params['transf']['input_size']
    if enc_input_size != trx_size:
        inp_reshape = PerTransTransf(trx_size, enc_input_size)
        p = torch.nn.Sequential(p, inp_reshape)

    e = TransformerSeqEncoder(enc_input_size, params['transf'])
    h = scoring_head(enc_input_size, params['head'])

    m = torch.nn.Sequential(p, e, h)
    return m


def model_by_type(model_type):
    model = {
        'avg': trx_avg_model,
        'avg2': trx_avg2_model,
        'rnn': rnn_model,
        'rnn-shuffle': rnn_shuffle_model,
        'skip-rnn2': skip_rnn2_model,
        'transf': transformer_model,
    }[model_type]
    return model


def freeze_layers(model):
    for p in model.parameters():
        p.requires_grad = False


def create_head_layers(params, seq_encoder=None):
    from torch.nn import Linear, BatchNorm1d, ReLU, Sigmoid, LogSoftmax, Dropout
    from dltranz.custom_layers import Squeeze, CombinedTargetHeadFromRnn, TargetHeadFromAggFeatures, DummyHead
    from dltranz.seq_encoder.utils import NormEncoder

    layers = []
    _locals = locals()
    for l_name, l_params in params['head_layers']:
        l_params = {k: int(v.format(**_locals)) if type(v) is str else v
                    for k, v in l_params.items()}

        cls = _locals.get(l_name, None)
        layers.append(cls(**l_params))
    return torch.nn.Sequential(*layers)


class Head(torch.nn.Module):
    r"""Head for the sequence encoder

    Parameters
    ----------
         input_size: int
            input size
         use_norm_encoder: bool. Default: False
            whether to use normalization layers before the head
         use_batch_norm: bool. Default: False.
            whether to use BatchNorm.
         hidden_layers_sizes: List[int]. Default: None.
            sizes of linear layers. If None without additional linear layers. Default = None,
         objective: str. Default: None.
            Options: None, 'classification', 'regression'. Default = None.
         num_classes: int. Default: 1.
            The number of classed in classification problem. Default correspond to binary classification.

     """
    def __init__(self,
                 input_size: int,
                 use_norm_encoder: bool = False,
                 use_batch_norm: bool = False,
                 hidden_layers_sizes: List[int] = None,
                 objective: str = None,
                 num_classes: int = 1):
        super().__init__()
        layers = []

        if use_norm_encoder:
            layers.append(NormEncoder())

        if use_batch_norm:
            layers.append(BatchNorm1d(input_size))

        if hidden_layers_sizes is not None:
            layers_size = [input_size] + list(hidden_layers_sizes)
            for size_in, size_out in zip(layers_size[:-1], layers_size[1:]):
                layers.append(Linear(size_in, size_out))
                layers.append(ReLU())
                if use_batch_norm:
                    layers.append(BatchNorm1d(size_out))

        if objective == 'classification':
            if num_classes == 1:
                h = Sequential(Linear(input_size, num_classes), Sigmoid(), Squeeze())
            else:
                h = Sequential(Linear(input_size, num_classes), LogSoftmax(dim=1))
            layers.append(h)

        elif objective == 'regression':
            h = Sequential(Linear(input_size, 1), Squeeze())
            layers.append(h)

        elif objective is not None:
            raise AttributeError(f"unknown objective {objective}. Supported: classification and regression")

        self.model = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)



# coding: utf-8
"""

"""
import functools
import operator

from torch.utils.data import DataLoader

from dltranz.data_load import padded_collate, ConvertingTrxDataset, DropoutTrxDataset, ProcessDataset
from dltranz.metric_learn.dataset.infinite_dataset import InfiniteDataset
from dltranz.metric_learn.dataset.splitting_dataset import SplittingDataset, SeveralSplittingsDataset
from dltranz.metric_learn.dataset import split_strategy
from dltranz.metric_learn.dataset.infinite_loader import InfiniteBatchSampler
from dltranz.metric_learn.dataset.target_enumerator_dataset import TargetEnumeratorDataset
from dltranz.metric_learn.dataset.complex_target_dataset import ComplexTargetDataset
from dltranz.metric_learn.dataset.preload_dataset import PreloadDataset, PreloadDataLoader


def nested_list_to_flat_with_collate(collate_fn):
    def _func(batch):
        batch = functools.reduce(operator.iadd, batch)
        return collate_fn(batch)

    return _func


def collate_splitted_rows(batch):
    # add Y and flatten samples
    batch = functools.reduce(operator.iadd, batch)
    return padded_collate(batch)


def create_train_data_loader(data, conf):
    dataset = SplittingDataset(
        data,
        split_strategy.create(**conf['params.train.split_strategy'])
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    dataset = DropoutTrxDataset(
        dataset,
        trx_dropout=conf['params.train.trx_dropout'],
        seq_len=conf['params.train.max_seq_len'])

    data_loader = DataLoader(
        dataset,
        shuffle=True,
        collate_fn=collate_splitted_rows,
        num_workers=conf['params.train.num_workers']
    )
    return data_loader


def create_train_infinite_data_loader(conf, prepare_gen):
    dataset = InfiniteDataset(conf['dataset'], prepare_gen=prepare_gen,
                              max_file_read=conf.get('dataset.max_file_read', None))
    valid_data = dataset.skip_first(conf['params.valid.read_size'])  # skip validation area

    dataset = SplittingDataset(
        dataset,
        split_strategy.create(**conf['params.train.split_strategy'])
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    dataset = DropoutTrxDataset(dataset, trx_dropout=conf['params.train.trx_dropout'],
                                seq_len=conf['params.train.max_seq_len'])

    batch_sampler = InfiniteBatchSampler(
        epoch_size=conf['params.train.epoch_size'],
        batch_size=conf['params.train.batch_size'],
    )
    data_loader = DataLoader(
        dataset,
        batch_sampler=batch_sampler,
        collate_fn=collate_splitted_rows,
        num_workers=0
    )
    return data_loader, valid_data


def create_train_preloading_data_loader(conf, prepare_gen, sample_process_fun=None):
    preload_dataset = PreloadDataset(conf, prepare_gen=prepare_gen,
                                     max_file_read=conf.get('dataset.max_file_read', None))
    valid_data = preload_dataset.skip_first(conf['params.valid.read_size'])  # skip validation area

    dataset = SplittingDataset(
        preload_dataset,
        split_strategy.create(**conf['params.train.split_strategy'])
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=conf['params.train.trx_dropout'],
                                seq_len=conf['params.train.max_seq_len'])

    data_loader = PreloadDataLoader(
        dataset=dataset,
        shuffle=True,
        collate_fn=collate_splitted_rows,
        num_workers=conf['params.train'].get('num_workers', 0),
        batch_size=conf['params.train.batch_size'],
    )
    data_loader.preload_dataset = preload_dataset

    return data_loader, valid_data


def create_valid_data_loader(data, conf, sample_process_fun=None):
    dataset = SplittingDataset(
        data,
        split_strategy.create(**conf['params.valid.split_strategy'])
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=0.0, seq_len=conf['params.valid.max_seq_len'])

    data_loader = DataLoader(
        dataset,
        batch_size=conf['params.valid.batch_size'],
        shuffle=False,
        collate_fn=collate_splitted_rows,
        num_workers=conf['params.valid.num_workers'],
    )
    return data_loader


def create_train_complex_data_loader(conf, prepare_gen, sample_process_fun=None):
    preload_dataset = PreloadDataset(conf, prepare_gen=prepare_gen,
                                     max_file_read=conf.get('dataset.max_file_read', None))
    valid_data = preload_dataset.skip_first(conf['params.valid.read_size'])  # skip validation area

    split_strategies = [
        split_strategy.create(**split_strategy_params) for split_strategy_params in
        conf['params.train.split_strategies']
    ]
    split_counts = [
        split_strategy_params['split_count'] for split_strategy_params in conf['params.train.split_strategies']
    ]
    dataset = SeveralSplittingsDataset(
        preload_dataset,
        split_strategies
    )
    dataset = ComplexTargetDataset(dataset, split_counts)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=conf['params.train.trx_dropout'],
                                seq_len=conf['params.train.max_seq_len'])

    data_loader = PreloadDataLoader(
        dataset=dataset,
        shuffle=True,
        collate_fn=collate_splitted_rows,
        num_workers=conf['params.train'].get('num_workers', 0),
        batch_size=conf['params.train.batch_size'],
    )
    data_loader.preload_dataset = preload_dataset

    return data_loader, valid_data


def create_valid_complex_data_loader(data, conf, sample_process_fun=None):
    split_strategies = [
        split_strategy.create(**split_strategy_params) for split_strategy_params in
        conf['params.valid.split_strategies']
    ]
    split_counts = [
        split_strategy_params['split_count'] for split_strategy_params in conf['params.valid.split_strategies']
    ]
    dataset = SeveralSplittingsDataset(
        data,
        split_strategies
    )
    dataset = ComplexTargetDataset(dataset, split_counts)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=0.0, seq_len=conf['params.valid.max_seq_len'])

    data_loader = DataLoader(
        dataset,
        batch_size=conf['params.valid.batch_size'],
        shuffle=False,
        collate_fn=collate_splitted_rows,
        num_workers=conf['params.valid.num_workers'],
    )
    return data_loader



import torch
from torch.nn import BCELoss

from dltranz.custom_layers import Squeeze
from dltranz.lightning_modules.AbsModule import ABSModule
from dltranz.seq_to_target import EpochAuroc
from dltranz.seq_encoder.utils import AllStepsHead, FlattenHead


class RtdModule(ABSModule):
    def __init__(self, params):
        super().__init__(params)

        self._head = torch.nn.Sequential(
            AllStepsHead(
                torch.nn.Sequential(
                    torch.nn.Linear(self.seq_encoder.embedding_size, 1),
                    torch.nn.Sigmoid(),
                    Squeeze(),
                )
            ),
            FlattenHead(),
        )

    @property
    def metric_name(self):
        return 'valid_auroc'

    @property
    def is_requires_reduced_sequence(self):
        return False

    def get_loss(self):
        return BCELoss()

    def get_validation_metric(self):
        return EpochAuroc()

    def shared_step(self, x, y):
        y_h = self.seq_encoder(x)
        y_h = self._head(y_h)
        return y_h, y



import pytorch_lightning as pl
from dltranz.seq_to_target import SequenceToTarget
from .test_data_load import RandomEventData
from pyhocon import ConfigFactory
from dltranz.seq_to_target import LogAccuracy
import torch


def tst_params_rnn():
    params = {
        "data_module": {
            "train": {
                "batch_size": 32,
                "num_workers": 1,
                "trx_dropout": .1,
                "max_seq_len": 30
            },
            "valid": {
                "batch_size": 32,
                "num_workers": 1,
                "max_seq_len": 30
            },

        },
        "encoder_type": "rnn",
        'score_metric': 'auroc',
        "train": {
            "weight_decay": 0,
            "lr": 0.004,
            "loss": "bce",
            "random_neg": True,
            "batch_size": 32,
            "num_workers": 1,
            "trx_dropout": .1,
            "n_epoch": 1,
            "max_seq_len": 30
        },
        "valid": {
            "batch_size": 32,
            "num_workers": 1,
            "max_seq_len": 30
        },
        'rnn': {
            'trainable_starter': 'static',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        "trx_encoder": {
            "embeddings_noise": .1,
            "norm_embeddings": False,
            'embeddings': {
                'mcc_code': {'in': 21, 'out': 3},
                'trans_type': {'in': 11, 'out': 2},
            },
            'numeric_values': {'amount': 'log'}
        },
        'head_layers': [
            ["Linear", {"in_features": "{seq_encoder.embedding_size}", "out_features": 1}],
            ["Sigmoid", {}],
            ["Squeeze", {}],
        ],
        "lr_scheduler": {
            "step_size": 10,
            "step_gamma": 0.8
        }
    }
    return ConfigFactory.from_dict(params)


def tst_params_transf():
    params = {
        'model_type': 'transf',
        'transf': {
            'n_heads': 2,
            'dim_hidden': 16,
            'dropout': .1,
            'n_layers': 2,
            'max_seq_len': 200,
            'use_after_mask': True,
            'use_positional_encoding': True,
            'sum_output': True,
            'input_size': 16,
            'train_starter': True,
            'shared_layers': False,
            'use_src_key_padding_mask': False
        },
        "head": {
            "explicit_lengths": False,
        }
    }

    return ConfigFactory.from_dict(params).with_fallback(tst_params_rnn())


def test_train_loop_rnn():
    params = tst_params_rnn()

    model = SequenceToTarget(params)
    dl = RandomEventData(params['data_module'])
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)


def test_train_loop_transf():
    params = tst_params_transf()

    model = SequenceToTarget(params)
    dl = RandomEventData(params['data_module'])
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)


def test_accuracy_bin():
    acc = LogAccuracy()
    y_hat = torch.tensor([0.1, 0.4, 0.6, 0.8, 0.9])
    y = torch.tensor([0, 1, 0, 1, 0])
    acc(y_hat, y)
    assert acc.compute() == 0.4


def test_accuracy_mul():
    acc = LogAccuracy()
    y_hat = torch.log_softmax(torch.tensor([
        [-1, 2, 1],
        [1, 2, -1],
        [1, -2, 0],
        [1, 1, 2],
        [1, 1, 2],
    ]).float(), dim=1)
    y = torch.tensor([0, 1, 1, 2, 2])
    acc(y_hat, y)
    assert acc.compute() == 0.6



import logging
import os
from functools import partial

import numpy as np
import pandas as pd

import torch
import tqdm
from sklearn.metrics import roc_auc_score, accuracy_score
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm

from dltranz.metric_learn.read_processing import fit_features, add_ticks, fit_types
from dltranz.data_load import read_dataset_mthread
from dltranz.util import get_data_files, get_conf
from dltranz.data_load import create_validation_loader, ConvertingTrxDataset, TrxDataset
from dltranz.train import score_model

logger = logging.getLogger(__name__)


class ModelEnsemble(torch.nn.Module):
    def __init__(self, submodels):
        super(ModelEnsemble, self).__init__()
        self.models = torch.nn.ModuleList(submodels)

    def forward(self, *args):
        out = []
        for m in self.models:
            o = m(*args)
            if len(o.size()) == 1:
                o = o.view(-1, 1)
            out.append(o)
        out = torch.cat(out, dim=1)
        return out


def load_model(conf):
    if 'model' in conf['model_path']:
        path = conf['model_path.model']
        model = torch.load(path, map_location=torch.device("cpu"))
        logger.info(f'Model loaded from "{path}"')
        return model
    elif 'models' in conf['model_path']:
        raise NotImplementedError()

        path = conf['model_path.models']
        models = [torch.load(p, map_location=torch.device("cpu")) for p in path]
        logger.info(f'Models loaded from "{path}"')
        return ModelEnsemble(models)

    elif 'state_dict' in conf['model_path']:
        raise NotImplementedError()
    else:
        raise AttributeError(f'Not supported model_path: {conf["model_path"]}')


def read_dataset_all(conf, desc, preproc_gen):
    logger.info(f'Data loading ({desc})...')
    files = get_data_files(conf['dataset'])
    n_workers = conf['dataset.n_workers']
    rec_gen = read_dataset_mthread(files, n_workers, preproc_gen)
    data = list(tqdm(rec_gen))
    logger.info(f'Loaded {len(data)} rows from disk ({desc})')
    return data


def infer_part_of_data(part_num, part_data, columns, model, conf, lock_obj=None):
    if lock_obj:
        lock_obj.acquire()

    if part_num is None:
        logger.info(f'Start to score data ({len(part_data)} records)')
    else:
        logger.info(f'Start to score {part_num} part of data ({len(part_data)} records)')

    if conf['dataset.preprocessing.add_seq_len'] and 'seq_len' not in columns:
        columns.append('seq_len')  # change list object

    valid_ds = ConvertingTrxDataset(TrxDataset(part_data))
    valid_loader = create_validation_loader(valid_ds, conf['params.valid'])

    _, pred = score_model(model, valid_loader, conf['params'])

    if conf['params.device'] != 'cpu':
        with torch.cuda.device(conf['params.device']):
            torch.cuda.empty_cache()
        logger.info('torch.cuda.empty_cache()')
    if lock_obj:
        lock_obj.release()

    if len(pred.shape) == 1:
        pred = pred.reshape(-1, 1)

    df_scores_cols = [f'v{i:003d}' for i in range(pred.shape[1])]
    df_scores = pd.DataFrame(pred, columns=df_scores_cols)

    df_labels = pd.DataFrame(({k: v for k, v in rec.items() if k in columns}
                              for rec in part_data))
    for col in df_labels:
        df_scores[col] = df_labels[col]
    df_scores = df_scores.reindex(columns=df_labels.columns.tolist() + df_scores_cols)
    logger.info(f'df_scores examples: {df_scores.shape}:')
    return df_scores


def infer_iterable(part_num, valid_loader, columns, model, conf, lock_obj=None):
    """
    The list of difference with `dltranz.metric_learn.inference_tools.infer_part_of_data`:
    1. Iterable dataset can'not provide the same item order.
        So you can'not get the same row order during two passes over the data
        (fist - embedding calculation, second - labels collecting, see `columns`)
        This why we put `id` field to `y` and get synchronised `true, outputs` from `score_model`,
        where `true` is target column and `outputs` is embeddings.
        This way requires only one pass over the dataset, and this way is faster than `infer_part_of_data`.
        But you are limited in `columns` feature: only scalar int ids are supported.
    2. Minor differences because of iterable_dataset has no len.
    """
    if part_num is not None:
        logger.warning(f'`part_name` aren\'t supported')

    if lock_obj:
        lock_obj.acquire()

    logger.info(f'Start to score data (iterable)')

    if conf['dataset.preprocessing.add_seq_len'] and 'seq_len' not in columns:
        columns.append('seq_len')  # change list object

    ids, pred = score_model(model, valid_loader, conf['params'])

    if conf['params.device'] != 'cpu':
        with torch.cuda.device(conf['params.device']):
            torch.cuda.empty_cache()
        logger.info('torch.cuda.empty_cache()')
    if lock_obj:
        lock_obj.release()

    if len(pred.shape) == 1:
        pred = pred.reshape(-1, 1)
    df_scores_cols = [f'v{i:003d}' for i in range(pred.shape[1])]
    col_id = conf['dataset.col_id']

    df_scores = pd.concat([
        pd.DataFrame({col_id: ids}),
        pd.DataFrame(pred, columns=df_scores_cols),
        ], axis=1)

    logger.info(f'df_scores examples: {df_scores.shape}:')
    return df_scores


def save_scores(df_scores, part_num, output_conf):
    # output
    output_name = output_conf['path']
    output_format = output_conf['format']
    if output_format not in ('pickle', 'csv'):
        logger.warning(f'Format "{output_format}" is not supported. Used default "pickle"')
        output_format = 'pickle'

    if part_num is None:
        output_path = f'{output_name}.{output_format}'
    else:
        os.makedirs(output_conf['path'], exist_ok=True)
        output_path = f'{output_name}/{part_num:03}.{output_format}'

    if output_format == 'pickle':
        df_scores.to_pickle(output_path)
    elif output_format == 'csv':
        df_scores.to_csv(output_path, sep=',', header=True, index=False)
    else:
        raise AssertionError('Never happens')
    logger.info(f'{len(df_scores)} records saved to: "{output_path}"')


def score_part_of_data(part_num, part_data, columns, model, conf, lock_obj=None):
    if type(part_data) is list:
        df_scores = infer_part_of_data(part_num, part_data, columns, model, conf, lock_obj=lock_obj)
    elif isinstance(part_data, DataLoader):
        df_scores = infer_iterable(part_num, part_data, columns, model, conf, lock_obj=lock_obj)
    else:
        valid_ds = ConvertingTrxDataset(TrxDataset(part_data))
        valid_loader = create_validation_loader(valid_ds, conf['params.valid'])
        df_scores = infer_iterable(part_num, valid_loader, columns, model, conf, lock_obj=lock_obj)

    save_scores(df_scores, part_num, conf['output'])


def common_preprocessing(seq, conf):
    preprocessing_conf = conf['dataset.preprocessing']

    fill_target = preprocessing_conf['fill_target']
    add_seq_len = preprocessing_conf['add_seq_len']
    min_date = np.datetime64(preprocessing_conf['min_date']) if 'min_date' in preprocessing_conf else None
    max_date = np.datetime64(preprocessing_conf['max_date']) if 'max_date' in preprocessing_conf else None

    for rec in seq:
        if 'application_date' in rec:
            application_date = np.datetime64(rec['application_date'])
            rec['application_date'] = application_date
            if min_date is not None and application_date < min_date:
                continue
            if max_date is not None and application_date >= max_date:
                continue

        if fill_target == 'empty':
            rec['target'] = -1

        if add_seq_len:
            tranz_dates = rec['event_time']
            n_tranz = len(tranz_dates)
            rec['seq_len'] = n_tranz

        yield rec


def consumer_preprocess_gen(seq, conf, data_pre_filter=None, ticks_mode='application_date'):
    if data_pre_filter is not None:
        seq = data_pre_filter(seq, conf)

    seq = common_preprocessing(seq, conf)
    seq = fit_features(seq,
                       embeddings=conf['params.trx_encoder.embeddings'],
                       numeric_values=conf['params.trx_encoder.numeric_values'])
    if 'tick_params' in conf['params']:
        seq = add_ticks(seq, conf, mode=ticks_mode)
    seq = fit_types(seq,
                    embeddings=conf['params.trx_encoder.embeddings'],
                    numeric_values=conf['params.trx_encoder.numeric_values'])
    return seq


def main_single_part(args=None):
    conf = get_conf(args)

    model = load_model(conf)
    columns = conf['output.columns']

    valid_data = read_dataset_all(conf, 'valid', partial(consumer_preprocess_gen, conf=conf))
    score_part_of_data(None, valid_data, columns, model, conf)


def score_data(conf, y_true, y_predict):
    metric_name = conf['params.score_metric']
    if metric_name not in ('auroc', 'accuracy'):
        raise AttributeError(f'Unknown metric: "{metric_name}"')

    col_id = conf['output.columns'][0]

    model_type = conf['params.model_type']
    if model_type in ('rnn', 'cpc_model'):
        cnt_features = conf['params.rnn.hidden_size']
    else:
        raise AttributeError(f'Unknown model_type: "{model_type}"')

    y_predict = y_predict.set_index(col_id)
    y_true = pd.DataFrame([{col_id: rec[col_id], 'target': rec['target']} for rec in y_true])
    y_true = y_true.set_index(col_id).reindex(index=y_predict.index)
    df = y_predict.merge(y_true, on=col_id, how='left')
    if metric_name == 'auroc':
        score = roc_auc_score(df['target'], df.iloc[:, 0])
    if metric_name == 'accuracy':
        score = accuracy_score(df['target'], np.argmax(df[y_predict.columns].values, axis=1))
    return {
        metric_name: score,
        'cnt_samples': len(y_true),
        'cnt_features': cnt_features,
    }



import logging
from copy import deepcopy

import pytorch_lightning as pl
import torch
import numpy as np
from torchmetrics.functional.classification import auroc
import torchmetrics

from dltranz.loss import get_loss, cross_entropy, kl, mape_metric, mse_loss, r_squared
from dltranz.seq_encoder import create_encoder
from dltranz.train import get_optimizer, get_lr_scheduler
from dltranz.models import create_head_layers
from dltranz.trx_encoder import PaddedBatch
from collections import defaultdict


logger = logging.getLogger(__name__)


class EpochAuroc(torchmetrics.Metric):
    def __init__(self):
        super().__init__(compute_on_step=False)

        self.add_state('y_hat', default=[])
        self.add_state('y', default=[])

    def update(self, y_hat, y):
        self.y_hat.append(y_hat)
        self.y.append(y)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return auroc(y_hat, y.long())


class DistributionTargets(torchmetrics.Metric):
    def __init__(self, col_name):
        super().__init__(compute_on_step=False)

        self.add_state('y_hat', default=[])
        self.add_state('y', default=[])
        self.col_name = col_name
        self.sign = -1 if self.col_name == 'neg_sum' else 1

    def update(self, y_hat, y):
        y_hat = y_hat[self.col_name]
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        y = torch.tensor(np.array(y[self.col_name].tolist(), dtype='float64'), device=device)
        self.y_hat.append(y_hat)
        self.y.append(y)


class CrossEntropy(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return cross_entropy(y_hat, y)


class KL(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return kl(y_hat, y)


class MSE(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return mse_loss(y_hat, torch.log(self.sign * y[:, None] + 1))


class MAPE(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return mape_metric(self.sign * torch.exp(y_hat - 1), y[:, None])


class LogAccuracy(torchmetrics.Accuracy):
    def __init__(self, **params):
        super().__init__(**params)

        self.add_state('correct', default=torch.tensor([0.0]))
        self.add_state('total', default=torch.tensor([0.0]))

    def update(self, preds, target):
        if len(preds.shape) == 1:
            preds = (preds > 0.5).to(dtype=target.dtype)
        else:
            preds = torch.argmax(preds, dim=1)
        self.correct += torch.sum(preds == target)
        self.total += target.numel()

    def compute(self):
        return self.correct / self.total


class R_squared(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return r_squared(self.sign * torch.exp(y_hat - 1), y[:, None])


class SequenceToTarget(pl.LightningModule):
    def __init__(self, params, pretrained_encoder=None):
        super().__init__()
        self.train_update_n_steps = params.get('train_update_n_steps', None)

        self.metrics_test = defaultdict(list)  # here we accumulate metrics on each test end
        self.metrics_train = defaultdict(list)  # here we accumulate metrics on some train bathes (called outside)

        head_params = dict(params['head_layers']).get('CombinedTargetHeadFromRnn', None)
        self.pos, self.neg = (head_params.get('pos', True), head_params.get('neg', True)) if head_params else (0, 0)
        self.cols_ix = params.get('columns_ix', {'neg_sum': 0,
                                                 'neg_distribution': 1,
                                                 'pos_sum': 2,
                                                 'pos_distribution': 3})

        self.save_hyperparameters('params')
        self.loss = get_loss(params)

        if pretrained_encoder is not None:
            self._seq_encoder = deepcopy(pretrained_encoder)
            self._is_pretrained_encoder = True
        else:
            self._seq_encoder = create_encoder(params, is_reduce_sequence=True)
            self._is_pretrained_encoder = False
        self._head = create_head_layers(params, self._seq_encoder)

        # metrics
        d_metrics = {
            'auroc': EpochAuroc(),
            'accuracy': LogAccuracy(),
            'R2n': R_squared('neg_sum'),
            'MSEn': MSE('neg_sum'),
            'MAPEn': MAPE('neg_sum'),
            'R2p': R_squared('pos_sum'),
            'MSEp': MSE('pos_sum'),
            'MAPEp': MAPE('pos_sum'),
            'CEn': CrossEntropy('neg_distribution'),
            'CEp': CrossEntropy('pos_distribution'),
            'KLn': KL('neg_distribution'),
            'KLp': KL('pos_distribution')
        }
        params_score_metric = params['score_metric']
        if type(params_score_metric) is str:
            params_score_metric = [params_score_metric]
        metric_cls = [(name, d_metrics[name]) for name in params_score_metric]

        self.train_metrics = torch.nn.ModuleDict([(name, mc) for name, mc in metric_cls])
        self.valid_metrics = torch.nn.ModuleDict([(name, mc) for name, mc in deepcopy(metric_cls)])
        self.test_metrics = torch.nn.ModuleDict([(name, mc) for name, mc in deepcopy(metric_cls)])


    @property
    def seq_encoder(self):
        return self._seq_encoder

    def forward(self, x):
        x = self._seq_encoder(x)
        x = self._head(x)
        return x

    def training_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        loss = self.loss(y_h, y)
        self.log('loss', loss)
        if isinstance(x, PaddedBatch):
            self.log('seq_len', x.seq_lens.float().mean(), prog_bar=True)
        if self.train_update_n_steps and self.global_step % self.train_update_n_steps == 0:
            for name, mf in self.train_metrics.items():
                mf(y_h, y)
        return loss

    def validation_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        for name, mf in self.valid_metrics.items():
            mf(y_h, y)

    def validation_epoch_end(self, outputs):
        for name, mf in self.valid_metrics.items():
            self.log(f'val_{name}', mf.compute(), prog_bar=True)

    def test_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        for name, mf in self.test_metrics.items():
            mf(y_h, y)

    def test_epoch_end(self, outputs):
        for name, mf in self.test_metrics.items():
            value = mf.compute().item()
            self.log(f'test_{name}', value, prog_bar=True)
            self.metrics_test[name] += [value]

    def configure_optimizers(self):
        params = self.hparams.params
        if self._is_pretrained_encoder:
            optimizer = self.get_pretrained_optimizer()
        else:
            optimizer = get_optimizer(self, params)
        scheduler = get_lr_scheduler(optimizer, params)
        return [optimizer], [scheduler]

    def get_pretrained_optimizer(self):
        params = self.hparams.params

        if params['pretrained.lr'] == 'freeze':
            self._seq_encoder.freeze()
            logger.info('Created optimizer with frozen encoder')
            return get_optimizer(self, params)

        parameters = [
            {'params': self._seq_encoder.parameters(), 'lr': params['pretrained.lr']},
            {'params': self._head.parameters(), 'lr': params['train.lr']},
        ]
        logger.info('Created optimizer with two lr groups')

        return torch.optim.Adam(parameters, lr=params['train.lr'], weight_decay=params['train.weight_decay'])



import logging
import os
from collections import namedtuple
from functools import reduce
from glob import glob
from multiprocessing.pool import Pool
from operator import iadd

import lightgbm as lgb
import pandas as pd
import scipy.stats
import xgboost as xgb
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.metrics import roc_auc_score, accuracy_score
from sklearn.model_selection import StratifiedKFold

import dltranz.fastai.fastai_tools as fai
import dltranz.neural_automl.neural_automl_tools as node

logger = logging.getLogger(__name__)


def prepare_common_parser(parser, data_path, output_file):
    parser.add_argument('--n_workers', type=int, default=5)
    parser.add_argument('--cv_n_split', type=int, default=5)
    parser.add_argument('--data_path', type=os.path.abspath, default=data_path)
    parser.add_argument('--random_state', type=int, default=42)
    parser.add_argument('--model_seed', type=int, default=42)
    parser.add_argument('--add_baselines', action='store_true')
    parser.add_argument('--add_emb_baselines', action='store_true')
    parser.add_argument('--baseline_name')
    parser.add_argument('--models', nargs='+', default=['linear', 'lgb', 'xgb'])
    parser.add_argument('--embedding_file_names', nargs='*', default=[])
    parser.add_argument('--score_file_names', nargs='*', default=[])
    parser.add_argument('--output_file', type=os.path.abspath, default=output_file)
    parser.add_argument('--labeled_amount', type=int, default=-1)


def read_train_test(data_path, dataset_file, test_ids_file, col_id):
    target = pd.read_csv(os.path.join(data_path, dataset_file))
    test_ids = set(pd.read_csv(os.path.join(data_path, test_ids_file))[col_id].tolist())

    ix_test = target[col_id].isin(test_ids)

    logger.info(f'Train size: {(~ix_test).sum()} clients')
    logger.info(f'Test size: {ix_test.sum()} clients')

    return target[~ix_test].set_index(col_id), target[ix_test].set_index(col_id)


def drop_na_target(col_target, df, df_name):
    ix = df[col_target].isna()
    logger.info(f'Drop {ix.sum()} rows with NA values from "{df_name}". There are {(~ix).sum()} rows left')
    return df[~ix]


def filter_infrequent_target(col_target, df, df_name, keep_values):
    ix = df[col_target].isin(keep_values)
    logger.info(f'Drop {(~ix).sum()} infrequent rows from "{df_name}". There are {ix.sum()} rows left')
    return df[ix]


def get_folds(df, col_target, cv_n_split, random_state, labeled_amount=-1):
    folds = []
    if labeled_amount < 0: labeled_amount = len(df)  # semi-supervised setup. default = supervised
    skf = StratifiedKFold(n_splits=cv_n_split, random_state=random_state, shuffle=True)
    for i_train, i_test in skf.split(df, df[col_target]):
        folds.append((
            df.iloc[i_train[:labeled_amount]],
            df.iloc[i_test]
        ))
    return folds


KWParamsTrainAndScore = namedtuple('KWParamsTrainAndScore', [
    'name',
    'fold_n',
    'load_features_f',
    'model_type',
    'model_params',
    'scorer_name',
    'scorer',
    'col_target',
    'df_train',
    'df_valid',
    'df_test',
])


def train_and_score(kw_params: KWParamsTrainAndScore):
    log_process_id = f'{kw_params.name:20}:{kw_params.model_type:6}:{kw_params.fold_n:2}'

    try:
        logger.info(f'[{log_process_id}] Started')

        features = kw_params.load_features_f()

        y_train = kw_params.df_train[kw_params.col_target]
        y_valid = kw_params.df_valid[kw_params.col_target]
        y_test = kw_params.df_test[kw_params.col_target]

        train_features = [df.reindex(index=kw_params.df_train.index) for df in features]
        valid_features = [df.reindex(index=kw_params.df_valid.index) for df in features]
        test_features = [df.reindex(index=kw_params.df_test.index) for df in features]

        df_fn = [df.quantile(0.5) for df in train_features]
        df_norm = [df.max() - df.min() + 1e-5 for df in train_features]

        X_train = pd.concat([df.fillna(fn) / n for df, fn, n in zip(train_features, df_fn, df_norm)], axis=1)
        X_valid = pd.concat([df.fillna(fn) / n for df, fn, n in zip(valid_features, df_fn, df_norm)], axis=1)
        X_test = pd.concat([df.fillna(fn) / n for df, fn, n in zip(test_features, df_fn, df_norm)], axis=1)

        if kw_params.model_type == 'linear':
            if kw_params.model_params.get('objective') == 'regression':
                model = LinearRegression(**{k: v for k, v in kw_params.model_params.items() if k != 'objective'})
            else:
                model = LogisticRegression(**{k: v for k, v in kw_params.model_params.items() if k != 'objective'})
        elif kw_params.model_type == 'xgb':
            if kw_params.model_params.get('objective', 'classification').startswith('reg'):
                model = xgb.XGBRegressor(**kw_params.model_params)
            else:
                model = xgb.XGBClassifier(**kw_params.model_params)
        elif kw_params.model_type == 'lgb':
            if kw_params.model_params.get('objective') == 'regression':
                model = lgb.LGBMRegressor(**kw_params.model_params)
            else:
                model = lgb.LGBMClassifier(**kw_params.model_params)
        elif kw_params.model_type == 'tabnet':
            if kw_params.model_params.get('objective') == 'regression':
                from pytorch_tabnet.tab_model import TabNetRegressor
                model = TabNetRegressor(**kw_params.model_params['model_params'])
            else:
                from pytorch_tabnet.tab_model import TabNetClassifier
                model = TabNetClassifier(**kw_params.model_params['model_params'])
        elif kw_params.model_type in ('neural_automl', 'fastai'):
            pass
        else:
            raise NotImplementedError(f'Unknown model type {kw_params.model_type}')

        if kw_params.model_type in ['linear', 'xgb', 'lgb']:
            model.fit(X_train, y_train)
            score_valid = kw_params.scorer(model, X_valid, y_valid)
            score_test = kw_params.scorer(model, X_test, y_test)
        elif kw_params.model_type == 'neural_automl':
            score_valid = node.train_from_config(X_train.values,
                                                 y_train.values.astype('long'),
                                                 X_valid.values,
                                                 y_valid.values.astype('long'),
                                                 kw_params.model_params)
            score_test = -1

        elif kw_params.model_type == 'fastai':
            score_valid = fai.train_from_config(X_train.values,
                                                y_train.values.astype('long'),
                                                X_valid.values,
                                                y_valid.values.astype('long'),
                                                kw_params.model_params)
            '''score_valid = fai.train_tabular(X_train.values, 
                                               y_train.values.astype('long'), 
                                               X_valid.values, 
                                               y_valid.values.astype('long'),
                                               kw_params.model_params)'''
            score_test = -1

        if kw_params.model_type == 'tabnet':
            model.fit(
                X_train.values,
                y_train.values,
                eval_set=[(X_valid.values, y_valid.values)],
                **kw_params.model_params['fit_params']
            )
            score_valid = kw_params.scorer(model, X_valid.values, y_valid.values)
            score_test = kw_params.scorer(model, X_test.values, y_test.values)

        logger.info(
            ' '.join([
                f'[{log_process_id}]',
                f'Finished with {kw_params.scorer_name}',
                f'valid={score_valid:.4f},',
                f'test={score_test:.4f}'
            ]))
    except Exception as ex:
        logging.exception(f'[{log_process_id}]: exception\n{ex}', exc_info=True)
        score_valid = None
        score_test = None

    res = {
        'name': '_'.join([kw_params.model_type, kw_params.name]),
        'fold_n': kw_params.fold_n,
        f'oof_{kw_params.scorer_name}': score_valid,
        f'test_{kw_params.scorer_name}': score_test,
    }
    return res


def train_and_score_mean_by_key(kw_params: KWParamsTrainAndScore):
    """
    The differences with `dltranz.scenario_cls_tools.train_and_score`:
    - Duplicated keys allowed in test_df and valid_df index
    - Probabilities from `model.predict` aggregates by `df.index` key with `mean` aggfunc
    - Aggregation is slightly different for roc_auc and accuracy metrics.
        This why you should choose metric and aggregation func manually

    :param kw_params:
    :return:
    """
    log_process_id = f'{kw_params.name:20}:{kw_params.model_type:6}:{kw_params.fold_n:2}'

    try:
        logger.info(f'[{log_process_id}] Started')

        features = kw_params.load_features_f()[0]

        train_ds = pd.merge(features, kw_params.df_train[[kw_params.col_target]], left_index=True, right_index=True)
        valid_ds = pd.merge(features, kw_params.df_valid[[kw_params.col_target]], left_index=True, right_index=True)
        test_ds = pd.merge(features, kw_params.df_test[[kw_params.col_target]], left_index=True, right_index=True)

        y_train = train_ds[kw_params.col_target]
        y_valid = valid_ds[kw_params.col_target]
        y_test = test_ds[kw_params.col_target]

        X_train = train_ds.drop(columns=kw_params.col_target)
        X_valid = valid_ds.drop(columns=kw_params.col_target)
        X_test = test_ds.drop(columns=kw_params.col_target)

        df_fn = X_train.quantile(0.5)
        df_norm = X_train.max() - X_train.min() + 1e-5

        X_train = X_train.fillna(df_fn) / df_norm
        X_valid = X_valid.fillna(df_fn) / df_norm
        X_test = X_test.fillna(df_fn) / df_norm

        if kw_params.model_type == 'linear':
            if kw_params.model_params.get('objective') == 'regression':
                model = LinearRegression(**{k:v for k,v in kw_params.model_params.items() if k != 'objective'})
            else:
                model = LogisticRegression(**{k:v for k,v in kw_params.model_params.items() if k != 'objective'})
        elif kw_params.model_type == 'xgb':
            if kw_params.model_params.get('objective', 'classification').startswith('reg'):
                model = xgb.XGBRegressor(**kw_params.model_params)
            else:
                model = xgb.XGBClassifier(**kw_params.model_params)
        elif kw_params.model_type == 'lgb':
            if kw_params.model_params.get('objective') == 'regression':
                model = lgb.LGBMRegressor(**kw_params.model_params)
            else:
                model = lgb.LGBMClassifier(**kw_params.model_params)
        elif kw_params.model_type in ('neural_automl', 'fastai'):
            pass
        else:
            raise NotImplementedError(f'Unknown model type {kw_params.model_type}')

        if kw_params.model_type in ['linear', 'xgb', 'lgb']:
            model.fit(X_train, y_train)

            def _score_roc_auc(model, df, y):
                df = df.copy()
                df['predict'] = model.predict_proba(df)[:, 1]
                df['y'] = y
                df = df.groupby(level=0).mean()
                return roc_auc_score(df['y'], df['predict'])

            def _score_accuracy(model, df, y):
                df = df.copy()
                predicts = pd.DataFrame(model.predict_proba(df), index=df.index)
                predicts = predicts.groupby(level=0).mean()

                df['y'] = y
                df = df.groupby(level=0).mean()
                return accuracy_score(df['y'], predicts.reindex(index=df.index).idxmax(axis=1))

            # TODO: Choose score function here manually
            score_valid = _score_accuracy(model, X_valid, y_valid)
            score_test = _score_accuracy(model, X_test, y_test)
        elif kw_params.model_type == 'neural_automl':
            raise NotImplementedError()

        elif kw_params.model_type == 'fastai':
            raise NotImplementedError()

        logger.info(
            ' '.join([
                f'[{log_process_id}]',
                f'Finished with {kw_params.scorer_name}',
                f'valid={score_valid:.4f},',
                f'test={score_test:.4f}'
            ]))
    except Exception as ex:
        logging.exception(f'[{log_process_id}]: exception\n{ex}', exc_info=True)
        score_valid = None
        score_test = None

    res = {
        'name': '_'.join([kw_params.model_type, kw_params.name]),
        'fold_n': kw_params.fold_n,
        f'oof_{kw_params.scorer_name}': score_valid,
        f'test_{kw_params.scorer_name}': score_test,
    }
    return res


def group_stat_results(df, group_col_name, col_agg_metric=None, col_list_metrics=None, eps=1e-12):
    def values(x):
        return '[' + ' '.join([f'{i:.3f}' for i in sorted(x)]) + ']'

    def t_interval(x, p=0.95):
        n = len(x)
        s = x.std(ddof=1)

        return scipy.stats.t.interval(p, n - 1, loc=x.mean(), scale=(s + eps) / (n ** 0.5))

    def t_int_l(x, p=0.95):
        return t_interval(x, p)[0]

    def t_int_h(x, p=0.95):
        return t_interval(x, p)[1]

    metric_aggregates = []
    metric_names = []
    if col_agg_metric is not None:
        metric_aggregates.extend([
            df.groupby(group_col_name)[m_col].agg(['mean', t_int_l, t_int_h, 'std', values])
            for m_col in col_agg_metric
        ])
        metric_names.extend(col_agg_metric)
    if col_list_metrics is not None:
        metric_aggregates.extend([
            df.groupby(group_col_name)[m_col].agg([values])
            for m_col in col_list_metrics
        ])
        metric_names.extend(col_list_metrics)

    df_results = pd.concat(metric_aggregates, axis=1, keys=metric_names).sort_index()
    return df_results


class WPool:
    def __init__(self, processes):
        self.processes = processes
        self.pool = Pool(processes=processes) if self.processes > 0 else None

    def map(self, func, iterable):
        if self.pool is not None:
            return self.pool.map(func, iterable)
        else:
            return reduce(iadd, map(func, iterable))

    def imap_unordered(self, func, iterable):
        if self.pool is not None:
            return self.pool.imap_unordered(func, iterable)
        else:
            return reduce(iadd, map(func, iterable))


def expand_path(data_path, wc_paths):
    data_path = os.path.join(data_path, '')  # ensure `/` at the end

    embedding_file_names = []
    for path in wc_paths:
        for n_path in glob(data_path + path):
            embedding_file_names.append(n_path[len(data_path):])
    logger.info(f'Found {len(embedding_file_names)} embedding files: [{embedding_file_names}]')
    return embedding_file_names


import numpy as np
import torch


class DropDay:
    def __init__(self, event_time_name='event_time'):
        self.event_time_name = event_time_name

    def __call__(self, x):
        mask = self.get_perm_ix(x[self.event_time_name])
        new_x = {k: v[mask] for k, v in x.items()}
        return new_x

    @staticmethod
    def get_perm_ix(event_time):
        days = torch.unique(event_time, sorted=True)
        ix = np.random.choice(len(days), 1)[0]
        mask = event_time != days[ix]
        return mask



import logging

from .base import DataPreprocessor

logger = logging.getLogger(__name__)


class PysparkDataPreprocessor(DataPreprocessor):
    def __init__(self, col_id, cols_event_time, cols_category, cols_log_norm, cols_identity, cols_target, print_dataset_info=False):
        super().__init__(col_id, cols_event_time, cols_category, cols_log_norm, cols_identity, cols_target)
        self.print_dataset_info = print_dataset_info

    def _reset(self):
        """Reset internal data-dependent state of the preprocessor, if necessary.
        __init__ parameters are not touched.
        """
        # TODO: pyspark reset data-dependent state of the preprocessor
        pass

    def fit(self, dt, **params):
        """
        Parameters
        ----------
        dt : pandas.DataFrame with flat data

        Returns
        -------
        self : object
            Fitted preprocessor.
        """
        # Reset internal state before fitting
        self._reset()

        # TODO: pyspark fit

        return self

    def transform(self, dt, **params):
        self.check_is_fitted()

        # TODO: pyspark transformation

        return dt

    def check_is_fitted(self):
        # TODO: pyspark is preprocessor fitted check
        pass



import torch

from dltranz.seq_encoder.rnn_encoder import RnnEncoder
from dltranz.trx_encoder import PaddedBatch
from dltranz.seq_encoder.utils import LastStepEncoder


def get_data():
    return PaddedBatch(
        payload=torch.arange(4*5*8).view(4, 8, 5).float(),
        length=torch.tensor([4, 2, 6, 8])
    )


def test_shape():
    params = {
        'hidden_size': 6,
        'type': 'gru',
        'bidir': False,
        'trainable_starter': 'static',
    }
    model = RnnEncoder(5, params)

    x = get_data()

    out = model(x)
    print(out.payload.shape)


def test_last_step():
    params = {
        'hidden_size': 6,
        'type': 'gru',
        'bidir': False,
        'trainable_starter': 'static',
    }
    model = torch.nn.Sequential(RnnEncoder(5, params), LastStepEncoder())

    x = get_data()

    h = model(x)
    print(h.shape)



from pyhocon import ConfigFactory
from dltranz.seq_to_target import SequenceToTarget


class SeqToTargetDemo(SequenceToTarget):
    """pl.LightningModule for training CoLES embeddings
    Parameters
    ----------
    seq_encoder : torch.nn.Module
        sequence encoder
    encoder_lr : float. Default: 1e-4
        Learning rate for sequence encoder.
    in_features: int. Default: 256
        Dimension of embeddings from sequence encoder.
    out_features: int. Default: 4
        Dimension of output of this model (number of classes).
    head_lr: float. Default: 0.002
    	Learning rate for head of this model.
    weight_decay: float. Default: 0.0
        Weight decay for optimizer.
    lr_scheduler_step_size : int. Default: 5
        Period of learning rate decay.
    lr_scheduler_step_gamma: float. Default: 0.4
        Multiplicative factor of learning rate decay.
    """

    def __init__(self,
                 seq_encoder = None,
                 encoder_lr: float = 0.0001,
                 in_features: int = 256,
                 out_features: int = 4,
                 head_lr: float = 0.002,
                 weight_decay: float = 0.0,
                 lr_scheduler_step_size: int = 5,
		 lr_scheduler_step_gamma: float = 0.4):

        params = {
            'score_metric': ['accuracy'],

            'encoder_type': 'pretrained',
            'pretrained': {
                'pl_module_class': 'dltranz.lightning_modules.coles_module.CoLESModule',
                'lr': encoder_lr
            },

            'head_layers': [
              ['BatchNorm1d', {'num_features': in_features}],
              ['Linear', {"in_features": in_features, "out_features": out_features}],
              ['LogSoftmax', {'dim': 1}]
            ],

            'train': {
              'random_neg': 'false',
              'loss': 'NLLLoss',
              'lr': head_lr,
              'weight_decay': weight_decay,
            },
            'lr_scheduler': {
              'step_size': lr_scheduler_step_size,
              'step_gamma': lr_scheduler_step_gamma
            }
        }
        super().__init__(ConfigFactory.from_dict(params), seq_encoder)




import pytorch_lightning as pl

from pyhocon import ConfigFactory

from dltranz.lightning_modules.emb_module import EmbModule
from dltranz.models import Head
from dltranz.seq_encoder import SequenceEncoder
from tests.dltranz_tests.test_data_load import RandomEventData


def tst_params():
    params = {
        "data_module": {
            "train": {
                "num_workers": 1,
                "batch_size": 32,
                "trx_dropout": 0.01,
                "max_seq_len": 100,
            },
            "valid": {
                "batch_size": 16,
                "num_workers": 1,
                "max_seq_len": 100
            }
        }
    }

    params = ConfigFactory.from_dict(params)
    return params


def test_train_loop():
    seq_encoder = SequenceEncoder(
        category_features={'mcc_code': 21, 'trans_type': 11},
        numeric_features=["amount"],
    )
    head = Head(
        input_size=seq_encoder.embedding_size,
        use_norm_encoder=True
    )
    model = EmbModule(
        seq_encoder=seq_encoder,
        head=head,
    )

    params = tst_params()

    dl = RandomEventData(params['data_module'])
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)



import torch

from dltranz.seq_encoder import SequenceEncoder
from dltranz.trx_encoder import PaddedBatch


def get_data():
    payload = {'amount': torch.arange(4 * 10).view(4, 10).float(),
               'event_time': torch.arange(4 * 10).view(4, 10).float(),
               'mcc_code': torch.arange(4 * 10).view(4, 10),
               'tr_type': torch.arange(4 * 10).view(4, 10)
               }
    return PaddedBatch(
        payload=payload,
        length=torch.tensor([4, 2, 6, 8])
    )


def test_shape():
    model = SequenceEncoder(
        category_features={'mcc_code': 200, 'tr_type': 100},
        numeric_features=["amount"],
        rnn_hidden_size=48
    )

    x = get_data()

    out = model(x)
    assert isinstance(out, torch.Tensor) and out.shape == torch.Size([4, 48])



import torch
from torch.nn import BCELoss

from dltranz.lightning_modules.AbsModule import ABSModule
from dltranz.seq_to_target import EpochAuroc


class SentencePairsHead(torch.nn.Module):
    def __init__(self, base_model, embeds_dim, config):
        super().__init__()
        self.base_model = base_model
        # TODO: check batch_norm position
        self.head = torch.nn.Sequential(
            torch.nn.Linear(embeds_dim * 2, config['hidden_size'], bias=True),
            torch.nn.BatchNorm1d(config['hidden_size']),
            torch.nn.ReLU(),
            torch.nn.Dropout(config['drop_p']),
            # torch.nn.BatchNorm1d(config['hidden_size']),
            torch.nn.Linear(config['hidden_size'], 1),
            torch.nn.Sigmoid()
        )

    def forward(self, x):
        left, right = x
        x = torch.cat([self.base_model(left), self.base_model(right)], dim=1)
        return self.head(x).squeeze(-1)


class SopNspModule(ABSModule):
    def __init__(self, params):
        super().__init__(params)

        self._head = SentencePairsHead(self.seq_encoder, self.seq_encoder.embedding_size, params['head'])

    @property
    def metric_name(self):
        return 'valid_auroc'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def get_loss(self):
        return BCELoss()

    def get_validation_metric(self):
        return EpochAuroc()

    def shared_step(self, x, y):
        y_h = self._head(x)
        return y_h, y



"""
Creates data loaders for Next Sequence Prediction learning
Each record pass throw the sampler and splits into views of the main record
Each sample slits into left and right part.
Some right part are shuffled.
X - is pair of feature_arrays (left and right), sub-samples of main event sequences
Y - is left and right from she same sample.
    0 means left and right from different sample, 1 means left and right from the same sample

    Input:        -> Output
dict of arrays    ->  X,                                            Y
client_1          -> (client_1_smpl_1_left, client_1_smpl_1_right), 1
                     (client_1_smpl_2_left, client_1_smpl_2_right), 1
                     (client_1_smpl_3_left, client_3_smpl_1_right), 0
client_2          -> (client_2_smpl_1_left, client_2_smpl_3_right), 0
                     ...
"""
import logging
import random

import numpy as np
import pytorch_lightning as pl
import torch

from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from dltranz.data_load.augmentations.build_augmentations import build_augmentations
from dltranz.data_load.augmentations.sequence_pair_augmentation import sequence_pair_augmentation
from dltranz.data_load import IterableChain, augmentation_chain, padded_collate_wo_target
from dltranz.data_load.iterable_processing.category_size_clip import CategorySizeClip
from dltranz.data_load.iterable_processing.feature_filter import FeatureFilter
from dltranz.data_load.iterable_processing.iterable_shuffle import IterableShuffle
from dltranz.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from dltranz.data_load.list_splitter import ListSplitter
from dltranz.data_load.parquet_dataset import ParquetDataset, ParquetFiles
from dltranz.data_load.partitioned_dataset import PartitionedDataset, PartitionedDataFiles
from dltranz.metric_learn.dataset import split_strategy, nested_list_to_flat_with_collate
from dltranz.metric_learn.dataset.splitting_dataset import IterableSplittingDataset, MapSplittingDataset

logger = logging.getLogger(__name__)


def collate_nsp_pairs(batch):
    lefts = [left for left, _ in batch] * 2

    rights = [right for _, right in batch]
    rights_ = rights[:]
    random.shuffle(rights_)
    rights += rights_

    targets = torch.cat([
        torch.ones(len(batch), dtype=torch.int64),
        torch.zeros(len(batch), dtype=torch.int64),
    ])

    return (
        (
            padded_collate_wo_target(lefts),
            padded_collate_wo_target(rights)
        ),
        targets.float(),
    )


class NspDataModuleTrain(pl.LightningDataModule):
    def __init__(self, conf, pl_module):
        super().__init__()

        self._type = conf['type']
        assert self._type in ('map', 'iterable')

        self.setup_conf = conf['setup']
        self.train_conf = conf['train']
        self.valid_conf = conf['valid']

        self.col_id = self.setup_conf['col_id']
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self._train_ids = None
        self._valid_ids = None

    def prepare_data(self):
        if 'dataset_files' in self.setup_conf:
            self.setup_iterable_files()
        elif 'dataset_parts' in self.setup_conf:
            self.setup_iterable_parts()
        else:
            raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')

        if self._type == 'map':
            self.setup_map()

    def setup_iterable_files(self):
        if self.setup_conf['split_by'] == 'files':
            data_files = ParquetFiles(self.setup_conf['dataset_files.data_path']).data_files

            splitter = ListSplitter(
                data_files,
                valid_size=self.setup_conf['valid_size'],
                seed=self.setup_conf['valid_split_seed'],
            )
            logger.info(f'Prepared splits: '
                        f'{len(splitter.train)} files in train, {len(splitter.valid)} files in valid')

            self.train_dataset = ParquetDataset(
                splitter.train,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = ParquetDataset(
                splitter.valid,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )

        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def setup_iterable_parts(self):
        data_files = PartitionedDataFiles(**self.setup_conf['dataset_parts'])

        if self.setup_conf['split_by'] == 'hash_id':
            splitter = ListSplitter(
                data_files.hash_parts,
                valid_size=self.setup_conf['valid_size'],
                seed=self.setup_conf['valid_split_seed'],
            )
            logger.info(f'Prepared splits: '
                        f'{len(data_files.dt_parts)} parts in dt_train, {len(data_files.dt_parts)} parts in dt_valid, '
                        f'{len(splitter.train)} parts in hash_train, {len(splitter.valid)} parts in hash_valid')

            self.train_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.train, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.valid, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )
        elif self.setup_conf['split_by'] == 'rows':
            raise NotImplementedError("Split by rows aren't supported for partitioned dataset")
        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def build_iterable_processing(self, part):
        if part == 'train':
            yield SeqLenFilter(min_seq_len=self.train_conf['min_seq_len'])

        yield FeatureFilter(keep_feature_names=self.category_names)
        yield CategorySizeClip(self.category_max_size)

        if self._type == 'iterable':
            # all processing in single chain
            if part == 'train':
                yield IterableShuffle(self.train_conf['buffer_size'])

            if part == 'train':
                split_strategy_conf = self.train_conf['split_strategy']
            else:
                split_strategy_conf = self.valid_conf['split_strategy']
            yield IterableSplittingDataset(split_strategy.create(**split_strategy_conf),
                                           self.build_augmentations(part))

    def build_augmentations(self, part):
        if part == 'train':
            a_chain = build_augmentations(self.train_conf['augmentations'])
        else:
            a_chain = build_augmentations(self.valid_conf['augmentations'])
        return augmentation_chain(a_chain, sequence_pair_augmentation)

    def setup_map(self):
        self.train_dataset = list(tqdm(iter(self.train_dataset)))
        logger.info(f'Loaded {len(self.train_dataset)} for train')
        self.valid_dataset = list(tqdm(iter(self.valid_dataset)))
        logger.info(f'Loaded {len(self.valid_dataset)} for valid')

        self.train_dataset = MapSplittingDataset(
            base_dataset=self.train_dataset,
            splitter=split_strategy.create(**self.train_conf['split_strategy']),
            a_chain=self.build_augmentations('train'),
        )
        self.valid_dataset = MapSplittingDataset(
            base_dataset=self.valid_dataset,
            splitter=split_strategy.create(**self.valid_conf['split_strategy']),
            a_chain=self.build_augmentations('valid'),
        )

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_nsp_pairs),
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf['num_workers'],
            batch_size=self.train_conf['batch_size'],
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_nsp_pairs),
            num_workers=self.valid_conf['num_workers'],
            batch_size=self.valid_conf['batch_size'],
        )



class InfiniteBatchSampler:
    def __init__(self, epoch_size, batch_size):
        self.epoch_size = epoch_size
        self.batch_size = batch_size

        self._current_epoch = 0

    def __iter__(self):
        start_pos = self._current_epoch * self.epoch_size
        current_indexes = range(start_pos, start_pos + self.epoch_size)

        # TODO: Можно добавить shuffle, если синхронизировать его с InfiniteDataset

        batch = []
        for ix in current_indexes:
            batch.append(ix)
            if len(batch) >= self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0:
            yield batch

        self._current_epoch += 1

    def __len__(self):
        return (self.epoch_size + self.batch_size - 1) // self.batch_size



from itertools import combinations

import numpy as np
import torch
import torch.nn.functional as F

from .metric import outer_pairwise_distance


class PairSelector:
    """
    Implementation should return indices of positive pairs and negative pairs that will be passed to compute
    Contrastive Loss
    return positive_pairs, negative_pairs
    """

    def __init__(self):
        pass

    def get_pairs(self, embeddings, labels):
        raise NotImplementedError


class AllPositivePairSelector(PairSelector):
    """
    Discards embeddings and generates all possible pairs given labels.
    If balance is True, negative pairs are a random sample to match the number of positive samples
    """

    def __init__(self, balance=True):
        super(AllPositivePairSelector, self).__init__()
        self.balance = balance

    def get_pairs(self, embeddings, labels):
        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        n = labels.size(0)
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)
        negative_pairs = torch.triu((x != 0).int(), diagonal=1).nonzero(as_tuple=False)

        if self.balance:
            negative_pairs = negative_pairs[torch.randperm(len(negative_pairs))[:len(positive_pairs)]]

        return positive_pairs, negative_pairs


class HardNegativePairSelector(PairSelector):
    """
    Generates all possible possitive pairs given labels and
         neg_count hardest negative example for each example
    """

    def __init__(self, neg_count=1):
        super(HardNegativePairSelector, self).__init__()
        self.neg_count = neg_count

    def get_pairs(self, embeddings, labels):
        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        n = labels.size(0)
        x = labels.expand(n, n) - labels.expand(n, n).t()

        # positive pairs
        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)
        
        # hard negative minning
        mat_distances = outer_pairwise_distance(embeddings.detach())  # pairwise_distance

        upper_bound = int((2 * n) ** 0.5) + 1
        mat_distances = ((upper_bound - mat_distances) * (x != 0).type(
            mat_distances.dtype))  # filter: get only negative pairs

        values, indices = mat_distances.topk(k=self.neg_count, dim=0, largest=True)
        negative_pairs = torch.stack([
            torch.arange(0, n, dtype=indices.dtype, device=indices.device).repeat(self.neg_count),
            torch.cat(indices.unbind(dim=0))
        ]).t()

        return positive_pairs, negative_pairs


class DistanceWeightedPairSelector(PairSelector):
    """
    Distance Weighted Sampling
    
    "Sampling Matters in Deep Embedding Learning", ICCV 2017
    https://arxiv.org/abs/1706.07567
    code based on https://github.com/suruoxi/DistanceWeightedSampling

    Generates pairs correspond to distances
    
    parameters
    ----------
    batch_k: int
        number of images per class

    Inputs:
        data: input tensor with shape (batch_size, embed_dim)
            Here we assume the consecutive batch_k examples are of the same class.
            For example, if batch_k = 5, the first 5 examples belong to the same class,
            6th-10th examples belong to another class, etc.
    Outputs:
        a_indices: indicess of anchors
        x[a_indices]
        x[p_indices]
        x[n_indices]
        xxx

    """

    def __init__(self, batch_k, cutoff=0.5, nonzero_loss_cutoff=1.4, normalize=False):
        super(DistanceWeightedPairSelector, self).__init__()
        self.batch_k = batch_k
        self.cutoff = cutoff
        self.nonzero_loss_cutoff = nonzero_loss_cutoff
        self.normalize = normalize

    def get_pairs(self, x, labels):
        k = self.batch_k
        n, d = x.shape
        distance = outer_pairwise_distance(x.detach())
        distance = distance.clamp(min=self.cutoff)
        log_weights = ((2.0 - float(d)) * distance.log() - (float(d - 3) / 2) * torch.log(
            torch.clamp(1.0 - 0.25 * (distance * distance), min=1e-8)))

        if self.normalize:
            log_weights = (log_weights - log_weights.min()) / (log_weights.max() - log_weights.min() + 1e-8)

        weights = torch.exp(log_weights - torch.max(log_weights))

        device = x.device
        weights = weights.to(device)

        mask = torch.ones_like(weights)
        for i in range(0, n, k):
            mask[i:i + k, i:i + k] = 0

        mask_uniform_probs = mask.double() * (1.0 / (n - k))

        weights = weights * mask * ((distance < self.nonzero_loss_cutoff).float()) + 1e-8
        weights_sum = torch.sum(weights, dim=1, keepdim=True)
        weights = weights / weights_sum

        a_indices = []
        p_indices = []
        n_indices = []

        np_weights = weights.cpu().numpy()
        for i in range(n):
            block_idx = i // k

            if weights_sum[i] != 0:
                n_indices += np.random.choice(n, k - 1, p=np_weights[i]).tolist()
            else:
                n_indices += np.random.choice(n, k - 1, p=mask_uniform_probs[i]).tolist()
            for j in range(block_idx * k, (block_idx + 1) * k):
                if j != i:
                    a_indices.append(i)
                    p_indices.append(j)

        positive_pairs = [[a, p] for a, p in zip(a_indices, p_indices)]
        negative_pairs = [[a, n] for a, n in zip(a_indices, n_indices)]

        return torch.LongTensor(positive_pairs).to(device), torch.LongTensor(negative_pairs).to(device)


class TripletSelector:
    """
    Implementation should return indices of anchors, positive and negative samples
    return np array of shape [N_triplets x 3]
    """

    def __init__(self):
        pass

    def get_triplets(self, embeddings, labels):
        raise NotImplementedError


class AllTripletSelector(TripletSelector):
    """
    Returns all possible triplets
    May be impractical in most cases
    """

    def __init__(self):
        super(AllTripletSelector, self).__init__()

    def get_triplets(self, embeddings, labels):
        np_labels = labels.cpu().data.numpy()
        triplets = []
        for label in set(np_labels):
            label_mask = (np_labels == label)
            label_indices = np.where(label_mask)[0]
            if len(label_indices) < 2:
                continue
            negative_indices = np.where(np.logical_not(label_mask))[0]
            anchor_positives = list(combinations(label_indices, 2))  # All anchor-positive pairs

            # Add all negatives for all positive pairs
            temp_triplets = [[anchor_positive[0], anchor_positive[1], neg_ind] for anchor_positive in anchor_positives
                             for neg_ind in negative_indices]
            triplets += temp_triplets

        return torch.LongTensor(np.array(triplets)).to(labels.device)


class RandomNegativeTripletSelector(TripletSelector):
    """
        Generate triplets with all positive pairs and random negative example for each anchor
    """

    def __init__(self, neg_count=1):
        super(RandomNegativeTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 1 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)
        anchor_labels = labels[positive_pairs[:, 0]]

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        negative_pairs = (x == 0).type(embeddings.dtype)
        negative_pairs_prob = (negative_pairs.t() / negative_pairs.sum(dim=1)).t()
        negative_pairs_rand = torch.multinomial(negative_pairs_prob, 1)

        triplets = torch.cat([positive_pairs, negative_pairs_rand], dim=1)

        return triplets


class HardTripletSelector(TripletSelector):
    """
        Generate triplets with all positive pairs and the neg_count hardest negative example for each anchor
    """

    def __init__(self, neg_count=1):
        super(HardTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)

        anchor_embed = embeddings[positive_pairs[:, 0]].detach()
        anchor_labels = labels[positive_pairs[:, 0]]

        # pos_embed = embeddings[positive_pairs[:,0]].detach()

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        mat_distances = outer_pairwise_distance(anchor_embed, embeddings.detach())  # pairwise_distance anchors x all

        upper_bound = int((2 * n) ** 0.5) + 1
        mat_distances = ((upper_bound - mat_distances) * (x == 0).type(
            mat_distances.dtype))  # filter: get only negative pairs

        values, indices = mat_distances.topk(k=self.neg_count, dim=1, largest=True)

        triplets = torch.cat([
            positive_pairs.repeat(self.neg_count, 1),
            torch.cat(indices.unbind(dim=0)).view(-1, 1)
        ], dim=1)

        return triplets


class SemiHardTripletSelector(TripletSelector):
    """
        Generate triplets with semihard sampling strategy

        "FaceNet: A Unified Embedding for Face Recognition and Clustering", CVPR 2015
        https://arxiv.org/abs/1503.03832
    """

    def __init__(self, neg_count=1):
        super(SemiHardTripletSelector, self).__init__()
        self.neg_count = neg_count

    def get_triplets(self, embeddings, labels):
        n = labels.size(0)

        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()

        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        m = positive_pairs.size(0)

        anchor_embed = embeddings[positive_pairs[:, 0]].detach()
        anchor_labels = labels[positive_pairs[:, 0]]

        pos_embed = embeddings[positive_pairs[:, 1]].detach()

        D_ap = F.pairwise_distance(anchor_embed, pos_embed)

        # construct matrix x (size m x n), such as x_ij == 1 <==> anchor_labels[i] == labels[j]
        x = (labels.expand(m, n) == anchor_labels.expand(n, m).t())

        mat_distances = outer_pairwise_distance(anchor_embed, embeddings.detach())  # pairwise_distance anchors x all

        neg_mat_distances = mat_distances * (x == 0).type(mat_distances.dtype)  # filter: get only negative pairs

        # negatives_outside: smallest D_an where D_an > D_ap.
        upper_bound = int((2 * n) ** 0.5) + 1
        negatives_outside = (upper_bound - neg_mat_distances) * \
                            (neg_mat_distances > D_ap.expand(n, m).t()).type(neg_mat_distances.dtype)
        values, negatives_outside = negatives_outside.topk(k=1, dim=1, largest=True)

        # negatives_inside: largest D_an
        values, negatives_inside = neg_mat_distances.topk(k=1, dim=1, largest=True)

        # whether exist negative n, such that D_an > D_ap.
        semihard_exist = ((neg_mat_distances > D_ap.expand(n, m).t()).sum(dim=1) > 0).view(-1, 1)

        negatives_indeces = torch.where(semihard_exist, negatives_outside, negatives_inside)

        triplets = torch.cat([positive_pairs, negatives_indeces], dim=1)

        return triplets


def get_sampling_strategy(params):
    if params['train.sampling_strategy'] is None:
        sampling_strategy = None

    elif params['train.sampling_strategy'] == 'AllPositivePair':
        sampling_strategy = AllPositivePairSelector()

    elif params['train.sampling_strategy'] == 'HardNegativePair':
        kwargs = {
            'neg_count': params.get('train.neg_count', None),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        sampling_strategy = HardNegativePairSelector(**kwargs)

    elif params['train.sampling_strategy'] == 'AllTriplets':
        sampling_strategy = AllTripletSelector()

    elif params['train.sampling_strategy'] == 'RandomNegativeTriplets':
        sampling_strategy = RandomNegativeTripletSelector()

    elif params['train.sampling_strategy'] == 'HardTriplets':
        kwargs = {
            'neg_count': params.get('train.neg_count', None),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        sampling_strategy = HardTripletSelector(**kwargs)

    elif params['train.sampling_strategy'] == 'SemiHardTriplets':
        sampling_strategy = SemiHardTripletSelector()

    elif params['train.sampling_strategy'] == 'DistanceWeightedPair':
        kwargs = {
            'batch_k': params['train.n_samples_from_class'],
            'cutoff': params.get('train.cutoff', None),
            'nonzero_loss_cutoff': params.get('train.nonzero_loss_cutoff', None),
            'normalize': params.get('train.normalize', None),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        sampling_strategy = DistanceWeightedPairSelector(**kwargs)

    else:
        raise AttributeError(f'wrong sampling_strategy "{params["train.sampling_strategy"]}"')

    return sampling_strategy



import torch


class AbsSeqEncoder(torch.nn.Module):
    def __init__(self, params, is_reduce_sequence):
        super().__init__()

        self.params = params
        self._is_reduce_sequence = is_reduce_sequence

    @property
    def is_reduce_sequence(self):
        return self._is_reduce_sequence

    @is_reduce_sequence.setter
    def is_reduce_sequence(self, value):
        self._is_reduce_sequence = value

    @property  
    def category_max_size(self):
        raise NotImplementedError()

    @property
    def category_names(self):
        raise NotImplementedError()

    @property
    def embedding_size(self):
        raise NotImplementedError()

    def freeze(self):
        for p in self.parameters():
            p.requires_grad = False



import numpy as np


class SeqLenLimit:
    def __init__(self, max_seq_len, strategy='tail'):
        self.max_seq_len = max_seq_len
        self.strategy = strategy

        assert strategy in ('tail', 'head', 'random')

    def __call__(self, x):
        seq_len = len(next(iter(x.values())))

        idx = self.get_idx(seq_len)
        new_x = {k: v[idx] for k, v in x.items()}
        return new_x

    def get_idx(self, seq_len):
        ix = np.arange(seq_len)
        if self.strategy == 'tail':
            return ix[-self.max_seq_len:]
        elif self.strategy == 'head':
            return ix[:self.max_seq_len]
        elif self.strategy == 'random':
            if seq_len <= self.max_seq_len:
                return ix
            else:
                max_start_pos = seq_len - self.max_seq_len
                start_pos = np.random.choice(max_start_pos, 1)[0]
                return ix[start_pos:start_pos + self.max_seq_len]



import logging
from glob import glob
from itertools import islice

from tqdm.autonotebook import tqdm

from dltranz.data_load import read_spark_dataset_gen, read_dataset_mthread
from torch.utils.data import Dataset, DataLoader

from dltranz.util import get_data_files, cycle_block_iterator

logger = logging.getLogger(__name__)


class PreloadDataset(Dataset):
    # TODO: можно добавить чтение заранее, чтобы подгрузка новых данных не начиналась в середине эпохи
    def __init__(self, params, prepare_gen, max_file_read=None, progress=False):
        self.dataset_params = params['dataset']
        self.prepare_gen = prepare_gen
        self.max_file_read = max_file_read
        self.n_workers = self.dataset_params.get('n_workers', 1)
        self.file_batch_size = self.dataset_params.get('file_batch_size', self.n_workers)
        if self.max_file_read is not None and self.n_workers > 1:
            self.n_workers = 1
            logger.warning(f'max_file_read mode is active. n_workers reduced to 1')
        self.progress = progress

        self._file_list = get_data_files(self.dataset_params)
        self._prepare_file_iterator()
        self.data = []
        self.train_epoch_size = params['params.train.epoch_size']
        self.n_epoch = -1

    def __len__(self):
        return self.train_epoch_size

    def __getitem__(self, item):
        return self.data[item]

    def _prepare_file_iterator(self):
        self.file_iterator = cycle_block_iterator(
            iter(range(len(self._file_list))),
            self.file_batch_size
        )

    def _read_next_files(self):
        file_real_num = next(self.file_iterator)
        files = [self._file_list[i] for i in file_real_num]

        gen = read_dataset_mthread(files, self.n_workers, self.prepare_gen)
        gen = tqdm(gen, desc=f'Read from files {file_real_num}/{len(self._file_list)}',
                   disable=not self.progress)
        if self.max_file_read is not None:
            gen = islice(gen, self.max_file_read)

        logger.info(f'Reading files {file_real_num}/{len(self._file_list)} ...')
        data = [x for x in gen]

        logger.info(f'Loaded {len(data)} rows from {len(files)} files')
        self.data += data

        return files

    def prepare_epoch(self):
        self.n_epoch += 1
        if self.n_epoch > 0:
            self.data = self.data[self.train_epoch_size:]
        while len(self.data) < self.train_epoch_size:
            self._read_next_files()
        logger.info(f'Epoch {self.n_epoch}, {len(self.data)} rows in buffer')

    def skip_first(self, size):
        readed_files = []
        while size > len(self.data):
            readed_files += self._read_next_files()

        skipped_data = self.data[:size]
        self.data = self.data[size:]
        self._file_list = [x for x in self._file_list if x not in readed_files]
        self._prepare_file_iterator()

        return skipped_data


class PreloadDataLoader(DataLoader):
    def prepare_epoch(self):
        self.preload_dataset.prepare_epoch()



import pickle

import torch

from dltranz.data_load import padded_collate, TrxDataset
from dltranz.trx_encoder import PaddedBatch, TrxEncoder, TrxMeanEncoder
from .test_data_load import gen_trx_data


def test_padded_batch_mask_tensor():
    data = PaddedBatch(torch.randn(4, 5, 3), torch.tensor([2, 5, 1, 3]))
    out = data.seq_len_mask
    exp = torch.tensor([
        [1, 1, 0, 0, 0],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0]
    ]).long()
    torch.testing.assert_close(out, exp)


def test_padded_batch_mask_dict():
    data = PaddedBatch({'col1': torch.randn(4, 5), 'col2': torch.randn(4, 5)}, torch.tensor([4, 5, 1, 3]))
    out = data.seq_len_mask
    exp = torch.tensor([
        [1, 1, 1, 1, 0],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0]
    ]).long()
    torch.testing.assert_close(out, exp)


def test_simple():
    x, y = padded_collate(TrxDataset(gen_trx_data([4, 3, 2])))

    params = {
        "norm_embeddings": False,
        "embeddings_noise": .1,
        'embeddings': {
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        'numeric_values': {'amount': 'log'}
    }

    te = TrxEncoder(params)

    e = te(x)

    assert e.payload.shape == (3, 4, 5)


def test_pickle():
    params = {
        "norm_embeddings": False,
        "embeddings_noise": .1,
        'embeddings': {
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        'numeric_values': {'amount': 'log'}
    }

    te = TrxEncoder(params)
    out = pickle.dumps(te)
    assert len(out) > 0


def test_mean_encoder():
    x, y = padded_collate(TrxDataset(gen_trx_data([4, 3, 2])))

    params = {
        "embeddings_noise": .1,
        'embeddings': {
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        'numeric_values': {'amount': 'log'}
    }

    te = TrxMeanEncoder(params)

    e = te(x)

    assert e.shape == (3, 33)



import numpy as np


class DropoutTrx:
    def __init__(self, trx_dropout):
        self.trx_dropout = trx_dropout

    def __call__(self, x):
        seq_len = len(next(iter(x.values())))

        idx = self.get_idx(seq_len)
        new_x = {k: v[idx] for k, v in x.items()}
        return new_x

    def get_idx(self, seq_len):
        if self.trx_dropout > 0 and seq_len > 0:
            idx = np.random.choice(seq_len, size=int(seq_len * (1 - self.trx_dropout)+1), replace=False)
            return np.sort(idx)
        else:
            return np.arange(seq_len)



from torch.utils.data import Dataset, DataLoader

from dltranz.metric_learn.dataset.infinite_loader import InfiniteBatchSampler


class TestInfiniteDataset(Dataset):
    def __getitem__(self, ix):
        return 1000 + ix


def test_sampler():
    data_loader = DataLoader(
        dataset=TestInfiniteDataset(),
        batch_sampler=InfiniteBatchSampler(epoch_size=5, batch_size=2),
        num_workers=0,
    )
    out_data = []
    for epoch in range(3):
        for batch in data_loader:
            out_data.append(batch.tolist())

    expected_out = [
        [1000, 1001],
        [1002, 1003],
        [1004],
        [1005, 1006],
        [1007, 1008],
        [1009],
        [1010, 1011],
        [1012, 1013],
        [1014],
    ]
    assert expected_out == out_data



import json
import logging

import pytorch_lightning as pl
import numpy as np
import ast
from embeddings_validation.file_reader import TargetFile, ID_TYPE_MAPPING
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from sklearn.model_selection import train_test_split
import random
import glob

from dltranz.data_load import padded_collate, padded_collate_distribution_target, IterableChain, IterableAugmentations
from dltranz.data_load.augmentations.build_augmentations import build_augmentations
from dltranz.data_load.data_module.map_augmentation_dataset import MapAugmentationDataset
from dltranz.data_load.iterable_processing.category_size_clip import CategorySizeClip
from dltranz.data_load.iterable_processing.feature_filter import FeatureFilter
from dltranz.data_load.iterable_processing.feature_type_cast import FeatureTypeCast
from dltranz.data_load.iterable_processing.id_filter import IdFilter
from dltranz.data_load.iterable_processing.iterable_shuffle import IterableShuffle
from dltranz.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from dltranz.data_load.iterable_processing.target_join import TargetJoin
from dltranz.data_load.parquet_dataset import ParquetFiles, ParquetDataset


logger = logging.getLogger(__name__)


class ClsDataModuleTrain(pl.LightningDataModule):
    def __init__(self, conf, pl_module, fold_id):
        super().__init__()

        self.fold_id = fold_id
        self._type = conf['type']
        assert self._type in ('map', 'iterable')

        self.distribution_targets_task = dict(conf).get('distribution_targets_task')
        self.y_function = int if not self.distribution_targets_task else lambda x: \
                                                    np.array(ast.literal_eval(x), dtype=object)
        self.setup_conf = conf['setup']
        self.train_conf = conf['train']
        self.valid_conf = conf['valid']
        self.test_conf = conf.get('test', self.valid_conf)

        self.col_id = self.setup_conf['col_id']
        self.col_id_dtype = {
            'str': str,
            'int': int,
        }[self.setup_conf['col_id_dtype']]
        self.col_target = self.setup_conf['col_target']
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self.test_dataset = None
        self._train_targets = None
        self._valid_targets = None
        self._test_targets = None

        self._fold_info = None

    def prepare_data(self):
        if 'dataset_files' in self.setup_conf:
            self.setup_iterable_files()
        elif 'dataset_parts' in self.setup_conf:
            self.setup_iterable_parts()
        else:
            raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')

        if self._type == 'map':
            self.setup_map()

    def setup_iterable_files(self):
        if self.setup_conf.get('split_by', None) == 'embeddings_validation':
          
            if (self.setup_conf.get('use_files_partially', None)):
                n_train = len(glob.glob(self.setup_conf['dataset_files.train_data_path'] + "/*.parquet"))
                ixes = list(range(n_train))
                train_ixes, test_ixes = train_test_split(ixes, test_size=int(n_train * (1 - self.setup_conf['train_part'])), shuffle=True)
                train_data_files = ParquetFiles(self.setup_conf['dataset_files.train_data_path'], train_ixes).data_files
                if self.setup_conf['same_file_for_test']:
                    test_ixes = random.sample(test_ixes, int(n_train * self.setup_conf['test_part']))
                    test_data_files = ParquetFiles(self.setup_conf['dataset_files.train_data_path'], test_ixes).data_files
                else:
                    n_test = len(glob.glob(self.setup_conf['dataset_files.test_data_path'] + "/*.parquet"))
                    test_ixes = random.sample(ixes, int(n_test * self.setup_conf['test_part']))
                    test_data_files = ParquetFiles(self.setup_conf['dataset_files.test_data_path'], test_ixes).data_files
            else:
                train_data_files = ParquetFiles(self.setup_conf['dataset_files.train_data_path']).data_files
                test_data_files = ParquetFiles(self.setup_conf['dataset_files.test_data_path']).data_files

            self.read_external_splits()
            self.train_dataset = ParquetDataset(
                train_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = ParquetDataset(
                train_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )
            self.test_dataset = ParquetDataset(
                test_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('test')),
                shuffle_files=False,
            )

        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def read_external_splits(self):
        if self.setup_conf['split_by'] == 'embeddings_validation':
            with open(self.setup_conf['fold_info'], 'r') as f:
                self._fold_info = json.load(f)

        current_fold = self._fold_info[self.fold_id]
        self._train_targets = TargetFile.load(current_fold['train']['path'])
        self._valid_targets = TargetFile.load(current_fold['valid']['path'])
        self._test_targets = TargetFile.load(current_fold['test']['path'])

        labeled_amount = self.train_conf.get('labeled_amount', None)
        if labeled_amount is not None:
            _len_orig = len(self._train_targets)
            if type(labeled_amount) is float:
                labeled_amount = int(_len_orig * labeled_amount)
            self._train_targets.df = self._train_targets.df.iloc[:labeled_amount]
            logger.info(f'Reduced train amount from {_len_orig} to {len(self._train_targets)}')

    def build_iterable_processing(self, part):
        yield FeatureTypeCast({self.col_id: self.col_id_dtype})

        if 'dataset_files' in self.setup_conf and self.setup_conf['split_by'] == 'embeddings_validation':
            if part == 'train':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._train_targets.df[self.col_id].values.tolist())
            elif part == 'valid':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._valid_targets.df[self.col_id].values.tolist())
            elif part == 'test':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._test_targets.df[self.col_id].values.tolist())
            else:
                raise AttributeError(f'Unknown part: {part}')

        if part == 'train':
            yield SeqLenFilter(min_seq_len=self.train_conf['min_seq_len'])

        if part == 'train':
            yield TargetJoin(self.col_id, self._train_targets.df.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        elif part == 'valid':
            yield TargetJoin(self.col_id, self._valid_targets.df.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        elif part == 'test':
            yield TargetJoin(self.col_id, self._test_targets.df.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        else:
            raise AttributeError(f'Unknown part: {part}')

        yield FeatureFilter(keep_feature_names=self.category_names)
        yield CategorySizeClip(self.category_max_size)

        if self._type == 'iterable':
            # all processing in single chain
            if part == 'train':
                yield IterableShuffle(self.train_conf['buffer_size'])

            yield IterableAugmentations(self.build_augmentations(part))

    def build_augmentations(self, part):
        if part == 'train':
            return build_augmentations(self.train_conf['augmentations'])
        elif part == 'valid':
            return build_augmentations(self.valid_conf['augmentations'])
        elif part == 'test':
            return build_augmentations(self.test_conf['augmentations'])

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_targets_task else padded_collate,
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf['num_workers'],
            batch_size=self.train_conf['batch_size'],
            drop_last=self.train_conf.get('drop_last', False)
        )

    def setup_map(self):
        self.train_dataset = list(tqdm(iter(self.train_dataset)))
        logger.info(f'Loaded {len(self.train_dataset)} for train')
        self.valid_dataset = list(tqdm(iter(self.valid_dataset)))
        logger.info(f'Loaded {len(self.valid_dataset)} for valid')
        self.test_dataset = list(tqdm(iter(self.test_dataset)))
        logger.info(f'Loaded {len(self.test_dataset)} for test')

        self.train_dataset = MapAugmentationDataset(
            base_dataset=self.train_dataset,
            a_chain=self.build_augmentations('train'),
        )
        self.valid_dataset = MapAugmentationDataset(
            base_dataset=self.valid_dataset,
            a_chain=self.build_augmentations('valid'),
        )
        self.test_dataset = MapAugmentationDataset(
            base_dataset=self.test_dataset,
            a_chain=self.build_augmentations('test'),
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_targets_task else padded_collate,
            num_workers=self.valid_conf['num_workers'],
            batch_size=self.valid_conf['batch_size'],
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.test_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_targets_task else padded_collate,
            num_workers=self.test_conf['num_workers'],
            batch_size=self.test_conf['batch_size'],
        )



from dltranz.lightning_modules.AbsModule import ABSModule
from dltranz.metric_learn.metric import BatchRecallTopPL
from dltranz.models import create_head_layers
from dltranz.metric_learn.losses import get_loss
from dltranz.metric_learn.sampling_strategies import get_sampling_strategy


class CoLESModule(ABSModule):
    def __init__(self, params):
        super().__init__(params)

        self._head = create_head_layers(params, self.seq_encoder)

    @property
    def metric_name(self):
        return 'recall_top_k'

    @property
    def is_requires_reduced_sequence(self):
        return True

    def get_loss(self):
        sampling_strategy = get_sampling_strategy(self.hparams.params)
        loss = get_loss(self.hparams.params, sampling_strategy)
        return loss

    def get_validation_metric(self):
        return BatchRecallTopPL(**self.hparams.params['validation_metric_params'])

    def shared_step(self, x, y):
        y_h = self(x)
        if self._head is not None:
            y_h = self._head(y_h)
        return y_h, y



import torch

from torch.utils.data import DataLoader
from tqdm import tqdm
from typing import List, Dict

from dltranz.data_load import IterableChain, padded_collate
from dltranz.data_load.filter_dataset import FilterDataset
from dltranz.data_load.iterable_processing.category_size_clip import CategorySizeClip
from dltranz.data_load.iterable_processing.feature_filter import FeatureFilter
from dltranz.data_load.iterable_processing.target_extractor import FakeTarget
from dltranz.train import score_model


def create_inference_dataloader(data: List[Dict],
                                num_workers: int = 0,
                                batch_size: int = 512,
                                category_names: List[str] = None,
                                category_max_size: Dict[str, int] = None):
    dataset = FilterDataset(
        data,
        post_processing=IterableChain(
            FakeTarget(),
            FeatureFilter(keep_feature_names=category_names),
            CategorySizeClip(category_max_size, 1))
    )
    dataset = list(tqdm(iter(dataset)))

    return DataLoader(
        dataset=dataset,
        collate_fn=padded_collate,
        shuffle=False,
        num_workers=num_workers,
        batch_size=batch_size,
    )


def get_embeddings(
        data: List[Dict],
        model: torch.nn.Module,
        num_workers: int = 0,
        batch_size: int = 512,
        category_names: List[str] = None,
        category_max_size: Dict[str, int] = None):
    """
    Parameters
    ----------
     data: List[Dict]
        data for inference
     model: torch.nn.Module
        pytorch model for inference
     num_workers: int. Default: 0.
        The number of workers for dataloader. 0 = single-process loader
     batch_size: int. Default: 512.
        The number of samples in each batch.
    category_names: List[str]. Default: None.
        The names of features (keys in data)
    category_max_size: Dict[str, int]. Default: None.
        The max value of categorical features (embedding matrix size)
    """
    embeds, _ = score_model(
        model=model,
        valid_loader=create_inference_dataloader(data, num_workers, batch_size, category_names, category_max_size)
    )
    return embeds



import logging
from bisect import bisect_right

import numpy as np
import torch

from tqdm import tqdm
from transformers import get_linear_schedule_with_warmup
from transformers.optimization import Adafactor, AdafactorSchedule

from dltranz.swa import SWA


logger = logging.getLogger(__name__)


def get_optimizer(model, params):
    """Returns optimizer

    :param model: model with his `model.named_parameters()`
    :param params: dict with options:
        ['train.lr']: `lr` for Adam optimizer
        ['train.weight_decay']: `weight_decay` for Adam optimizer
        ['train.optimiser_params']: (optional) list of tuples (par_name, options),
            each tuple define new parameter group.
            `par_name` is end of parameter name from `model.named_parameters()` for this parameter group
            'options' is dict with options for this parameter group
    :return:
    """
    optimiser_params = params.get('train.optimiser_params', None)
    if optimiser_params is None:
        parameters = model.parameters()
    elif optimiser_params.get('type', None) == 'Adafactor':
        optimizer = Adafactor(model.parameters(), scale_parameter=True, relative_step=True, warmup_init=True, lr=None)
        Adafactor(model.parameters(), scale_parameter=False, relative_step=False, warmup_init=False, lr=params['train.lr'])
        logger.info("Ada factor optimizer is used")
    elif optimiser_params.get('type', None) == 'SWA':
        optimizer = SWA(optimizer)
    else:
        parameters = []
        for par_name, options in optimiser_params.items():
            options = options.copy()
            options['params'] = [v for k, v in model.named_parameters() if k.startswith(par_name)]
            parameters.append(options)
        default_options = {
            'params': [v for k, v in model.named_parameters() if all(
                (not k.startswith(par_name) for par_name, options in optimiser_params.items())
            )]}
        parameters.append(default_options)
    optimizer = torch.optim.Adam(parameters, lr=params['train.lr'], weight_decay=params['train.weight_decay'])   

    return optimizer


class SchedulerWrapper(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, scheduler):
        self.scheduler = scheduler

    def __call__(self, *args, **kwargs):
        self.scheduler.step()

    def step(self, epoch=None):
        self.scheduler.step(epoch)

    @property
    def optimizer(self):
        return self.scheduler.optimizer


class ReduceLROnPlateauWrapper(torch.optim.lr_scheduler.ReduceLROnPlateau):
    def __init__(self, scheduler):
        self.scheduler = scheduler

    def __call__(self, metric_value, *args, **kwargs):
        self.scheduler.step(metric_value)

    def step(self, metric, epoch=None):
        self.scheduler.step(metric, epoch)

    @property
    def optimizer(self):
        return self.scheduler.optimizer


class MultiGammaScheduler(torch.optim.lr_scheduler.MultiStepLR):

    def __init__(self, optimizer, milestones, gammas, gamma=0.1, last_epoch=-1):
        self.gammas = gammas
        super(MultiGammaScheduler, self).__init__(optimizer, milestones, gamma, last_epoch)

    def get_lr(self):
        idx = bisect_right(self.milestones, self.last_epoch)
        gammas = self.gammas[:idx]
        gamma = np.prod(gammas)
        return [base_lr * gamma for base_lr in self.base_lrs]


def get_lr_scheduler(optimizer, params):
    if 'scheduler' in params:
        # TODO: check the this code branch
        if params['scheduler.current'] != '':
            scheduler_type = params['scheduler.current']

            scheduler_params = params[f'scheduler.{scheduler_type}']

            if scheduler_type == 'MultiGammaScheduler':
                scheduler = MultiGammaScheduler(optimizer,
                                         milestones=scheduler_params['milestones'],
                                         gammas=scheduler_params['gammas'],
                                         gamma=scheduler_params['gamma'],
                                         last_epoch=scheduler_params['last_epoch'])

            logger.info('MultiGammaScheduler used')

    elif params['lr_scheduler'].get('CosineAnnealing', False):
        T_max = params['train'].get('n_epoch', params['lr_scheduler.n_epoch'])
        eta_min = params['lr_scheduler'].get('eta_min', 0)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=T_max, eta_min=eta_min)
        logger.info('CosineAnnealingLR lr_scheduler used')
        wrapper = SchedulerWrapper

    elif params['lr_scheduler'].get('ReduceLROnPlateau', False):
        mode = params['lr_scheduler'].get('mode', 'max')
        factor = params['lr_scheduler'].get('factor', 0.1)
        patience = params['lr_scheduler'].get('patience', 10)
        threshold = params['lr_scheduler'].get('threshold', 0.001)
        min_lr = params['lr_scheduler'].get('min_lr', 1e-6)

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            mode=mode,
            factor=factor,
            patience=patience,
            threshold=threshold,
            threshold_mode='rel',
            min_lr=min_lr,
            verbose=True
        )
        logger.info('ReduceLROnPlateau lr_scheduler used')
        wrapper = ReduceLROnPlateauWrapper

    else:
        lr_step_size = params['lr_scheduler']['step_size']
        lr_step_gamma = params['lr_scheduler']['step_gamma']
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=lr_step_size, gamma=lr_step_gamma)
        logger.info('StepLR lr_scheduler used')
        wrapper = SchedulerWrapper

    # TODO: ReduceLROnPlateau + warmup
    if 'warmup' in params['lr_scheduler']:
        scheduler = get_linear_schedule_with_warmup(optimizer, 
           num_warmup_steps=params['lr_scheduler']["warmup"]['num_warmup_steps'], 
           num_training_steps=params['lr_scheduler']["warmup"]['num_training_steps'])
        logger.info('LR warmup used')
        # optimiser param groups are not supported with LRScheduler
        # create_lr_scheduler_with_warmup don't works with SchedulerWrapper
    else:
        scheduler = wrapper(scheduler)

    return scheduler


def score_model(model, valid_loader, params=None):
    """
      - extended valid_loader. input format: x, * in batch:
      - output: pred(x), * in score_model

    Returns:

    """
    if params is None:
        params = {}

    if torch.cuda.is_available():
        device = torch.device(params.get('device', 'cuda'))
    else:
        device = torch.device(params.get('device', 'cpu'))
    model.to(device)
    model.eval()

    outputs = []
    with torch.no_grad():
        for batch in tqdm(valid_loader, leave=False):
            x, *others = batch
            x = x.to(device)
            out = model(x)

            batch_output = [out.cpu().numpy(), *others]
            outputs.append(batch_output)

    outputs = zip(*outputs)
    outputs = (np.concatenate(l) for l in outputs)
    return outputs



import logging
from glob import glob
from itertools import islice

from tqdm.autonotebook import tqdm

from dltranz.data_load import read_spark_dataset_gen, read_dataset_mthread
from torch.utils.data import Dataset

from dltranz.util import get_data_files

logger = logging.getLogger(__name__)


class InfiniteDataset(Dataset):
    # TODO: можно добавить чтение заранее, чтобы подгрузка новых данных не начиналась в середине эпохи
    def __init__(self, dataset_params, prepare_gen, max_file_read=None, progress=False):
        self.dataset_params = dataset_params
        self.prepare_gen = prepare_gen
        self.max_file_read = max_file_read
        self.n_workers = dataset_params.get('n_workers', 1)
        self.file_batch_size = dataset_params.get('file_batch_size', self.n_workers)
        if self.max_file_read is not None and self.n_workers > 1:
            self.n_workers = 1
            logger.warning('max_file_read mode is active. n_workers reduced to 1')
        self.progress = progress

        self._file_list = get_data_files(dataset_params)
        self._next_file = 0

        self._current_buffer = None
        self._global_start_pos = 0
        self._global_next_file_pos = 0

    def __len__(self):
        raise AssertionError('Can not estimate size of infinite dataset')

    def __getitem__(self, item):
        if item >= self._global_next_file_pos:
            self._read_next_files()

        if item >= self._global_next_file_pos:
            raise IndexError(f"Can not go to future: "
                             f"{item} not in [{self._global_start_pos} : {self._global_next_file_pos}]")
        if item < self._global_start_pos:
            raise IndexError(f"Can not go to past: "
                             f"{item} not in [{self._global_start_pos} : {self._global_next_file_pos}]")

        return self._current_buffer[item - self._global_start_pos]

    def _read_next_files(self):
        file_log_num = [self._next_file + i for i in range(self.file_batch_size)]
        file_real_num = [i % len(self._file_list) for i in file_log_num]
        files = [self._file_list[i] for i in file_real_num]

        gen = read_dataset_mthread(files, self.n_workers, self.prepare_gen)
        gen = tqdm(gen, desc=f'Read from files {file_log_num}/{len(self._file_list)} (real file nums {file_real_num})',
                   disable=not self.progress)
        if self.max_file_read is not None:
            gen = islice(gen, self.max_file_read)

        logger.info(f'Reading files {file_log_num}/{len(self._file_list)} (real file nums {file_real_num}) ...')
        data = [x for x in gen]
        self._next_file += len(files)
        logger.info(f'Loaded {len(data)} rows from {len(files)} files')

        self._current_buffer = data
        self._global_start_pos = self._global_next_file_pos
        self._global_next_file_pos = self._global_start_pos + len(data)

    def skip_first(self, size):
        while size >= self._global_next_file_pos:
            self._read_next_files()

        buffer_new_start = size - self._global_start_pos
        skipped_data = self._current_buffer[:buffer_new_start]
        self._current_buffer = self._current_buffer[buffer_new_start:]
        self._global_start_pos = 0  # = _global_start_pos - size + buffer_new_start
        self._global_next_file_pos -= size
        return skipped_data



import numpy as np
import torch
import pytorch_lightning as pl
from dltranz.data_load import create_train_loader, create_validation_loader, TrxDataset


def gen_trx_data(lengths, target_share=.5):
    n = len(lengths)
    targets = (torch.rand(n) >= target_share).long()
    samples = list()
    for target, length in zip(targets, lengths):
        s = dict()
        s['trans_type'] = (torch.rand(length)*10 + 1).long()
        s['mcc_code'] = (torch.rand(length) * 20 + 1).long()
        s['amount'] = (torch.rand(length) * 1000 + 1).long()

        samples.append({'feature_arrays': s, 'target': target})

    return samples


class RandomEventData(pl.LightningDataModule):
    def __init__(self, params):
        super().__init__()
        self.hparams.update(params)

    def train_dataloader(self):
        test_data = TrxDataset(gen_trx_data((torch.rand(1000)*60+1).long()), y_dtype=np.int64)
        train_loader = create_train_loader(test_data, self.hparams['train'])
        return train_loader

    def val_dataloader(self):
        test_data = TrxDataset(gen_trx_data((torch.rand(200)*60+1).long()), y_dtype=np.int64)
        train_loader = create_validation_loader(test_data, self.hparams['valid'])
        return train_loader
